from datetime import timedelta
from throttled.asyncio import Throttled, rate_limiter
from throttled import Throttled as ThrottledSync
from throttled import rate_limiter as rate_limiter_sync

from nexustrader.constants import (
    AccountType,
    OrderStatus,
    PositionSide,
    OrderSide,
    TimeInForce,
    OrderType,
    KlineInterval,
    RateLimiter,
    RateLimiterSync,
    TriggerType,
)
from enum import Enum
from nexustrader.error import KlineSupportedError


class BybitOpType(Enum):
    AUTH = "auth"
    PING = "ping"
    PONG = "pong"
    ORDER_CREATE = "order.create"
    ORDER_AMEND = "order.amend"
    ORDER_CANCEL = "order.cancel"
    ORDER_CREATE_BATCH = "order.create-batch"
    ORDER_AMEND_BATCH = "order.amend-batch"
    ORDER_CANCEL_BATCH = "order.cancel-batch"


class BybitKlineInterval(Enum):
    MINUTE_1 = "1"
    MINUTE_3 = "3"
    MINUTE_5 = "5"
    MINUTE_15 = "15"
    MINUTE_30 = "30"
    HOUR_1 = "60"
    HOUR_2 = "120"
    HOUR_4 = "240"
    HOUR_6 = "360"
    HOUR_12 = "720"
    DAY_1 = "D"
    WEEK_1 = "W"
    MONTH_1 = "M"


class BybitAccountType(AccountType):
    SPOT = "SPOT"
    LINEAR = "LINEAR"
    INVERSE = "INVERSE"
    OPTION = "OPTION"
    SPOT_TESTNET = "SPOT_TESTNET"
    LINEAR_TESTNET = "LINEAR_TESTNET"
    INVERSE_TESTNET = "INVERSE_TESTNET"
    OPTION_TESTNET = "OPTION_TESTNET"
    UNIFIED = "UNIFIED"
    UNIFIED_TESTNET = "UNIFIED_TESTNET"
    LINEAR_MOCK = "LINEAR_MOCK"
    INVERSE_MOCK = "INVERSE_MOCK"
    SPOT_MOCK = "SPOT_MOCK"

    @property
    def exchange_id(self):
        return "bybit"

    @property
    def is_testnet(self):
        return self in {
            self.SPOT_TESTNET,
            self.LINEAR_TESTNET,
            self.INVERSE_TESTNET,
            self.OPTION_TESTNET,
            self.UNIFIED_TESTNET,
        }

    @property
    def ws_public_url(self):
        return WS_PUBLIC_URL[self]

    @property
    def ws_private_url(self):
        if self.is_testnet:
            return "wss://stream-testnet.bybit.com/v5/private"
        return "wss://stream.bybit.com/v5/private"

    @property
    def ws_api_url(self):
        if self.is_testnet:
            return "wss://stream-testnet.bybit.com/v5/trade"
        return "wss://stream.bybit.com/v5/trade"

    @property
    def is_spot(self):
        return self in {self.SPOT, self.SPOT_TESTNET}

    @property
    def is_linear(self):
        return self in {self.LINEAR, self.LINEAR_TESTNET}

    @property
    def is_inverse(self):
        return self in {self.INVERSE, self.INVERSE_TESTNET}

    @property
    def is_mock(self):
        return self in {self.LINEAR_MOCK, self.INVERSE_MOCK, self.SPOT_MOCK}

    @property
    def is_linear_mock(self):
        return self == self.LINEAR_MOCK

    @property
    def is_inverse_mock(self):
        return self == self.INVERSE_MOCK

    @property
    def is_spot_mock(self):
        return self == self.SPOT_MOCK


WS_PUBLIC_URL = {
    BybitAccountType.SPOT: "wss://stream.bybit.com/v5/public/spot",
    BybitAccountType.LINEAR: "wss://stream.bybit.com/v5/public/linear",
    BybitAccountType.INVERSE: "wss://stream.bybit.com/v5/public/inverse",
    BybitAccountType.OPTION: "wss://stream.bybit.com/v5/public/option",
    BybitAccountType.SPOT_TESTNET: "wss://stream-testnet.bybit.com/v5/public/spot",
    BybitAccountType.LINEAR_TESTNET: "wss://stream-testnet.bybit.com/v5/public/linear",
    BybitAccountType.INVERSE_TESTNET: "wss://stream-testnet.bybit.com/v5/public/inverse",
    BybitAccountType.OPTION_TESTNET: "wss://stream-testnet.bybit.com/v5/public/option",
}


class BybitBaseUrl(Enum):
    MAINNET_1 = "MAINNET_1"
    MAINNET_2 = "MAINNET_2"
    TESTNET = "TESTNET"
    NETHERLAND = "NETHERLAND"
    HONGKONG = "HONGKONG"
    TURKEY = "TURKEY"
    HAZAKHSTAN = "HAZAKHSTAN"

    @property
    def base_url(self):
        return REST_API_URL[self]


REST_API_URL = {
    BybitBaseUrl.MAINNET_1: "https://api.bybit.com",
    BybitBaseUrl.MAINNET_2: "https://api.bytick.com",
    BybitBaseUrl.TESTNET: "https://api-testnet.bybit.com",
    BybitBaseUrl.NETHERLAND: "https://api.bybit.nl",
    BybitBaseUrl.HONGKONG: "https://api.byhkbit.com",
    BybitBaseUrl.TURKEY: "https://api.bybit-tr.com",
    BybitBaseUrl.HAZAKHSTAN: "https://api.bybit.kz",
}


class BybitOrderSide(Enum):
    BUY = "Buy"
    SELL = "Sell"


class BybitOrderStatus(Enum):
    NEW = "New"
    REJECTED = "Rejected"
    PARTIALLY_FILLED = "PartiallyFilled"
    PARTIALLY_FILLED_CANCELED = "PartiallyFilledCanceled"
    FILLED = "Filled"
    CANCELED = "Cancelled"
    UNTRIGGERED = "Untriggered"
    TRIGGERED = "Triggered"
    DEACTIVATED = "Deactivated"


class BybitTimeInForce(Enum):
    GTC = "GTC"
    IOC = "IOC"
    FOK = "FOK"
    POST_ONLY = "PostOnly"


class BybitPositionIdx(Enum):
    FLAT = 0
    LONG = 1
    SHORT = 2

    def is_one_way_mode(self):
        return self == self.FLAT


class BybitPositionSide(Enum):
    FLAT = ""
    BUY = "Buy"
    SELL = "Sell"

    def parse_to_position_side(self) -> PositionSide:
        if self == self.FLAT:
            return PositionSide.FLAT
        elif self == self.BUY:
            return PositionSide.LONG
        elif self == self.SELL:
            return PositionSide.SHORT
        raise RuntimeError(f"Invalid position side: {self}")


class BybitOrderType(Enum):
    MARKET = "Market"
    LIMIT = "Limit"
    UNKNOWN = "Unknown"


class BybitProductType(Enum):
    SPOT = "spot"
    LINEAR = "linear"
    INVERSE = "inverse"
    OPTION = "option"

    @property
    def is_spot(self):
        return self == self.SPOT

    @property
    def is_linear(self):
        return self == self.LINEAR

    @property
    def is_inverse(self):
        return self == self.INVERSE

    @property
    def is_option(self):
        return self == self.OPTION


class BybitTriggerType(Enum):
    NONE = ""  # Default
    LAST_PRICE = "LastPrice"
    INDEX_PRICE = "IndexPrice"
    MARK_PRICE = "MarkPrice"


class BybitTriggerDirection(Enum):
    NONE = 0
    RISES_TO = 1  # Triggered when market price rises to triggerPrice
    FALLS_TO = 2


class BybitStopOrderType(Enum):
    NONE = ""  # Default
    UNKNOWN = "UNKNOWN"  # Classic account value
    TAKE_PROFIT = "TakeProfit"
    STOP_LOSS = "StopLoss"
    TRAILING_STOP = "TrailingStop"
    STOP = "Stop"
    PARTIAL_TAKE_PROFIT = "PartialTakeProfit"
    PARTIAL_STOP_LOSS = "PartialStopLoss"
    TPSL_ORDER = "tpslOrder"
    OCO_ORDER = "OcoOrder"  # Spot only
    MM_RATE_CLOSE = "MmRateClose"
    BIDIRECTIONAL_TPSL_ORDER = "BidirectionalTpslOrder"


class BybitEnumParser:
    _bybit_kline_interval_map = {
        BybitKlineInterval.MINUTE_1: KlineInterval.MINUTE_1,
        BybitKlineInterval.MINUTE_3: KlineInterval.MINUTE_3,
        BybitKlineInterval.MINUTE_5: KlineInterval.MINUTE_5,
        BybitKlineInterval.MINUTE_15: KlineInterval.MINUTE_15,
        BybitKlineInterval.MINUTE_30: KlineInterval.MINUTE_30,
        BybitKlineInterval.HOUR_1: KlineInterval.HOUR_1,
        BybitKlineInterval.HOUR_2: KlineInterval.HOUR_2,
        BybitKlineInterval.HOUR_4: KlineInterval.HOUR_4,
        BybitKlineInterval.HOUR_6: KlineInterval.HOUR_6,
        BybitKlineInterval.HOUR_12: KlineInterval.HOUR_12,
        BybitKlineInterval.DAY_1: KlineInterval.DAY_1,
        BybitKlineInterval.WEEK_1: KlineInterval.WEEK_1,
        BybitKlineInterval.MONTH_1: KlineInterval.MONTH_1,
    }

    _bybit_order_status_map = {
        BybitOrderStatus.NEW: OrderStatus.ACCEPTED,
        BybitOrderStatus.PARTIALLY_FILLED: OrderStatus.PARTIALLY_FILLED,
        BybitOrderStatus.PARTIALLY_FILLED_CANCELED: OrderStatus.CANCELED,
        BybitOrderStatus.FILLED: OrderStatus.FILLED,
        BybitOrderStatus.CANCELED: OrderStatus.CANCELED,
        BybitOrderStatus.TRIGGERED: OrderStatus.ACCEPTED,
        BybitOrderStatus.UNTRIGGERED: OrderStatus.PENDING,
        BybitOrderStatus.DEACTIVATED: OrderStatus.EXPIRED,
        BybitOrderStatus.REJECTED: OrderStatus.FAILED,
    }

    _bybit_position_side_map = {
        BybitPositionIdx.FLAT: PositionSide.FLAT,
        BybitPositionIdx.LONG: PositionSide.LONG,
        BybitPositionIdx.SHORT: PositionSide.SHORT,
    }

    _bybit_order_side_map = {
        BybitOrderSide.BUY: OrderSide.BUY,
        BybitOrderSide.SELL: OrderSide.SELL,
    }

    _bybit_order_time_in_force_map = {
        BybitTimeInForce.IOC: TimeInForce.IOC,
        BybitTimeInForce.GTC: TimeInForce.GTC,
        BybitTimeInForce.FOK: TimeInForce.FOK,
        BybitTimeInForce.POST_ONLY: TimeInForce.GTC,
    }

    _bybit_order_type_map = {
        BybitOrderType.MARKET: OrderType.MARKET,
        BybitOrderType.LIMIT: OrderType.LIMIT,
    }

    _bybit_trigger_type_map = {
        BybitTriggerType.LAST_PRICE: TriggerType.LAST_PRICE,
        BybitTriggerType.INDEX_PRICE: TriggerType.INDEX_PRICE,
        BybitTriggerType.MARK_PRICE: TriggerType.MARK_PRICE,
    }

    # Add reverse mapping dictionaries
    _order_status_to_bybit_map = {v: k for k, v in _bybit_order_status_map.items()}
    # Ensure CANCELED maps to the right status
    _order_status_to_bybit_map[OrderStatus.CANCELED] = BybitOrderStatus.CANCELED
    _position_side_to_bybit_map = {v: k for k, v in _bybit_position_side_map.items()}
    _order_side_to_bybit_map = {v: k for k, v in _bybit_order_side_map.items()}
    _time_in_force_to_bybit_map = {
        v: k for k, v in _bybit_order_time_in_force_map.items()
    }
    _time_in_force_to_bybit_map[TimeInForce.GTC] = BybitTimeInForce.GTC
    _order_type_to_bybit_map = {
        v: k for k, v in _bybit_order_type_map.items() if v is not None
    }
    _kline_interval_to_bybit_map = {
        v: k for k, v in _bybit_kline_interval_map.items() if v is not None
    }
    _trigger_type_to_bybit_map = {
        v: k for k, v in _bybit_trigger_type_map.items() if v is not None
    }

    @classmethod
    def parse_trigger_type(cls, trigger_type: BybitTriggerType) -> TriggerType:
        return cls._bybit_trigger_type_map[trigger_type]

    @classmethod
    def parse_kline_interval(cls, interval: BybitKlineInterval) -> KlineInterval:
        return cls._bybit_kline_interval_map[interval]

    # Add reverse parsing methods
    @classmethod
    def parse_order_status(cls, status: BybitOrderStatus) -> OrderStatus:
        return cls._bybit_order_status_map[status]

    @classmethod
    def parse_position_side(cls, side: BybitPositionIdx) -> PositionSide:
        return cls._bybit_position_side_map[side]

    @classmethod
    def parse_order_side(cls, side: BybitOrderSide) -> OrderSide:
        return cls._bybit_order_side_map[side]

    @classmethod
    def parse_time_in_force(cls, tif: BybitTimeInForce) -> TimeInForce:
        return cls._bybit_order_time_in_force_map[tif]

    @classmethod
    def parse_order_type(
        cls, order_type: BybitOrderType, time_in_force: TimeInForce | None = None
    ) -> OrderType:
        if time_in_force == BybitTimeInForce.POST_ONLY:
            return OrderType.POST_ONLY
        return cls._bybit_order_type_map[order_type]

    @classmethod
    def to_bybit_order_status(cls, status: OrderStatus) -> BybitOrderStatus:
        return cls._order_status_to_bybit_map[status]

    @classmethod
    def to_bybit_position_side(cls, side: PositionSide) -> BybitPositionIdx:
        return cls._position_side_to_bybit_map[side]

    @classmethod
    def to_bybit_order_side(cls, side: OrderSide) -> BybitOrderSide:
        return cls._order_side_to_bybit_map[side]

    @classmethod
    def to_bybit_time_in_force(cls, tif: TimeInForce) -> BybitTimeInForce:
        return cls._time_in_force_to_bybit_map[tif]

    @classmethod
    def to_bybit_order_type(cls, order_type: OrderType) -> BybitOrderType:
        return cls._order_type_to_bybit_map[order_type]

    @classmethod
    def to_bybit_kline_interval(cls, interval: KlineInterval) -> BybitKlineInterval:
        if interval not in cls._kline_interval_to_bybit_map:
            raise KlineSupportedError(
                f"Kline interval {interval} is not supported by Bybit"
            )
        return cls._kline_interval_to_bybit_map[interval]

    @classmethod
    def to_bybit_trigger_type(cls, trigger_type: TriggerType) -> BybitTriggerType:
        return cls._trigger_type_to_bybit_map[trigger_type]


class BybitRateLimiter(RateLimiter):
    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: dict[str, Throttled] = {
            "public": Throttled(
                quota=rate_limiter.per_duration(timedelta(seconds=5), limit=600),
                timeout=5 if enable_rate_limit else -1,
            ),
            "trade": Throttled(
                quota=rate_limiter.per_sec(20),
                timeout=1 if enable_rate_limit else -1,
            ),
            "position": Throttled(
                quota=rate_limiter.per_sec(50),
                timeout=1 if enable_rate_limit else -1,
            ),
            "account": Throttled(
                quota=rate_limiter.per_sec(50),
                timeout=1 if enable_rate_limit else -1,
            ),
            "ws/order": Throttled(
                quota=rate_limiter.per_sec(20),
                timeout=1 if enable_rate_limit else -1,
            ),
        }

    def __call__(self, rate_limit_type: str) -> Throttled:
        return self._throttled[rate_limit_type]


class BybitRateLimiterSync(RateLimiterSync):
    def __init__(self, enable_rate_limit: bool = True):
        self._throttled: dict[str, ThrottledSync] = {
            "public": ThrottledSync(
                quota=rate_limiter_sync.per_duration(timedelta(seconds=5), limit=600),
                timeout=5 if enable_rate_limit else -1,
            ),
            "trade": ThrottledSync(
                quota=rate_limiter_sync.per_sec(20),
                timeout=1 if enable_rate_limit else -1,
            ),
            "position": ThrottledSync(
                quota=rate_limiter_sync.per_sec(50),
                timeout=1 if enable_rate_limit else -1,
            ),
            "account": ThrottledSync(
                quota=rate_limiter_sync.per_sec(50),
                timeout=1 if enable_rate_limit else -1,
            ),
        }

    def __call__(self, rate_limit_type: str) -> ThrottledSync:
        return self._throttled[rate_limit_type]



import hmac
import hashlib
import msgspec
from typing import Any, Dict, List
from urllib.parse import urljoin, urlencode
from curl_cffi import requests
from decimal import Decimal
from nexustrader.base import ApiClient, RetryManager
from nexustrader.exchange.bybit.constants import (
    BybitBaseUrl,
    BybitRateLimiter,
    BybitRateLimiterSync,
)
from nexustrader.exchange.bybit.error import BybitError
from nexustrader.core.nautilius_core import hmac_signature, LiveClock
from nexustrader.exchange.bybit.schema import (
    BybitResponse,
    BybitOrderResponse,
    BybitPositionResponse,
    BybitOrderHistoryResponse,
    BybitOpenOrdersResponse,
    BybitWalletBalanceResponse,
    BybitKlineResponse,
    BybitIndexKlineResponse,
    BybitBatchOrderResponse,
    BybitTickersResponse,
)


class BybitApiClient(ApiClient):
    _limiter: BybitRateLimiter
    _limiter_sync: BybitRateLimiterSync

    def __init__(
        self,
        clock: LiveClock,
        api_key: str = None,
        secret: str = None,
        timeout: int = 10,
        testnet: bool = False,
        enable_rate_limit: bool = True,
        max_retries: int = 0,
        delay_initial_ms: int = 100,
        delay_max_ms: int = 800,
        backoff_factor: int = 2,
    ):
        """
        ### Testnet:
        `https://api-testnet.bybit.com`

        ### Mainnet:
        (both endpoints are available):
        `https://api.bybit.com`
        `https://api.bytick.com`

        ### Important:
        Netherland users: use `https://api.bybit.nl` for mainnet
        Hong Kong users: use `https://api.byhkbit.com` for mainnet
        Turkey users: use `https://api.bybit-tr.com` for mainnet
        Kazakhstan users: use `https://api.bybit.kz` for mainnet
        """

        super().__init__(
            clock=clock,
            api_key=api_key,
            secret=secret,
            timeout=timeout,
            rate_limiter=BybitRateLimiter(enable_rate_limit),
            rate_limiter_sync=BybitRateLimiterSync(enable_rate_limit),
            retry_manager=RetryManager(
                max_retries=max_retries,
                delay_initial_ms=delay_initial_ms,
                delay_max_ms=delay_max_ms,
                backoff_factor=backoff_factor,
                exc_types=(BybitError,),
                retry_check=lambda e: int(e.code)
                in [
                    429,
                    10000,
                    10016,
                    3400214,
                    170007,
                    177002,
                    500001,
                ],  # please refer to https://bybit-exchange.github.io/docs/v5/error
            ),
        )
        self._recv_window = 5000

        if testnet:
            self._base_url = BybitBaseUrl.TESTNET.base_url
        else:
            self._base_url = BybitBaseUrl.MAINNET_1.base_url

        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
        }

        if api_key:
            self._headers["X-BAPI-API-KEY"] = api_key

        self._msg_decoder = msgspec.json.Decoder()
        self._msg_encoder = msgspec.json.Encoder()
        self._response_decoder = msgspec.json.Decoder(BybitResponse)
        self._order_response_decoder = msgspec.json.Decoder(BybitOrderResponse)
        self._position_response_decoder = msgspec.json.Decoder(BybitPositionResponse)
        self._order_history_response_decoder = msgspec.json.Decoder(
            BybitOrderHistoryResponse
        )
        self._open_orders_response_decoder = msgspec.json.Decoder(
            BybitOpenOrdersResponse
        )
        self._wallet_balance_response_decoder = msgspec.json.Decoder(
            BybitWalletBalanceResponse
        )
        self._kline_response_decoder = msgspec.json.Decoder(BybitKlineResponse)
        self._index_kline_response_decoder = msgspec.json.Decoder(
            BybitIndexKlineResponse
        )
        self._batch_order_response_decoder = msgspec.json.Decoder(
            BybitBatchOrderResponse
        )
        self._tickers_response_decoder = msgspec.json.Decoder(BybitTickersResponse)

    def _generate_signature(self, payload: str) -> List[str]:
        timestamp = str(self._clock.timestamp_ms())

        param = str(timestamp) + self._api_key + str(self._recv_window) + payload
        hash = hmac.new(
            bytes(self._secret, "utf-8"), param.encode("utf-8"), hashlib.sha256
        )
        signature = hash.hexdigest()
        return [signature, timestamp]

    def _generate_signature_v2(self, payload: str) -> List[str]:
        timestamp = str(self._clock.timestamp_ms())
        param = f"{timestamp}{self._api_key}{self._recv_window}{payload}"
        signature = hmac_signature(self._secret, param)  # return hex digest string
        return [signature, timestamp]

    async def _fetch(
        self,
        method: str,
        base_url: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        return await self._retry_manager.run(
            name=f"{method} {endpoint}",
            func=self._fetch_async,
            method=method,
            base_url=base_url,
            endpoint=endpoint,
            payload=payload,
            signed=signed,
        )

    async def _fetch_async(
        self,
        method: str,
        base_url: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        self._init_session()

        url = urljoin(base_url, endpoint)
        payload = payload or {}

        payload_str = (
            urlencode(payload)
            if method == "GET"
            else self._msg_encoder.encode(payload).decode("utf-8")
        )

        headers = self._headers
        if signed:
            signature, timestamp = self._generate_signature_v2(payload_str)
            headers = {
                **headers,
                "X-BAPI-TIMESTAMP": timestamp,
                "X-BAPI-SIGN": signature,
                "X-BAPI-RECV-WINDOW": str(self._recv_window),
            }

        if method == "GET":
            url += f"?{payload_str}"
            payload_str = None

        try:
            self._log.debug(f"Request: {url} {payload_str}")
            response = await self._session.request(
                method=method,
                url=url,
                headers=headers,
                data=payload_str,
            )
            raw = response.content
            if response.status_code >= 400:
                raise BybitError(
                    code=response.status_code,
                    message=self._msg_decoder.decode(raw) if raw else None,
                )
            bybit_response: BybitResponse = self._response_decoder.decode(raw)
            if bybit_response.retCode == 0:
                return raw
            else:
                raise BybitError(
                    code=bybit_response.retCode,
                    message=bybit_response.retMsg,
                )
        except requests.exceptions.Timeout as e:
            self._log.error(f"Timeout {method} Url: {url} - {e}")
            raise
        except requests.exceptions.ConnectionError as e:
            self._log.error(f"Connection Error {method} Url: {url} - {e}")
            raise
        except requests.exceptions.HTTPError as e:
            self._log.error(f"HTTP Error {method} Url: {url} - {e}")
            raise
        except requests.exceptions.RequestException as e:
            self._log.error(f"Request Error {method} Url: {url} - {e}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} - {e}")
            raise

    def _fetch_sync(
        self,
        method: str,
        base_url: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        self._init_sync_session()

        url = urljoin(base_url, endpoint)
        payload = payload or {}

        payload_str = (
            urlencode(payload)
            if method == "GET"
            else self._msg_encoder.encode(payload).decode("utf-8")
        )

        headers = self._headers
        if signed:
            signature, timestamp = self._generate_signature_v2(payload_str)
            headers = {
                **headers,
                "X-BAPI-TIMESTAMP": timestamp,
                "X-BAPI-SIGN": signature,
                "X-BAPI-RECV-WINDOW": str(self._recv_window),
            }

        if method == "GET":
            url += f"?{payload_str}"
            payload_str = None

        try:
            self._log.debug(f"Request: {url} {payload_str}")
            response = self._sync_session.request(
                method=method,
                url=url,
                headers=headers,
                data=payload_str,
            )
            raw = response.content
            if response.status_code >= 400:
                raise BybitError(
                    code=response.status_code,
                    message=self._msg_decoder.decode(raw) if raw else None,
                )
            bybit_response: BybitResponse = self._response_decoder.decode(raw)
            if bybit_response.retCode == 0:
                return raw
            else:
                raise BybitError(
                    code=bybit_response.retCode,
                    message=bybit_response.retMsg,
                )
        except requests.exceptions.Timeout as e:
            self._log.error(f"Timeout {method} Url: {url} - {e}")
            raise
        except requests.exceptions.ConnectionError as e:
            self._log.error(f"Connection Error {method} Url: {url} - {e}")
            raise
        except requests.exceptions.HTTPError as e:
            self._log.error(f"HTTP Error {method} Url: {url} - {e}")
            raise
        except requests.exceptions.RequestException as e:
            self._log.error(f"Request Error {method} Url: {url} - {e}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} - {e}")
            raise

    async def post_v5_order_create(
        self,
        category: str,
        symbol: str,
        side: str,
        order_type: str,
        qty: Decimal,
        **kwargs,
    ) -> BybitOrderResponse:
        """
        https://bybit-exchange.github.io/docs/v5/order/create-order
        """
        endpoint = "/v5/order/create"
        payload = {
            "category": category,
            "symbol": symbol,
            "side": side,
            "orderType": order_type,
            "qty": str(qty),
            **kwargs,
        }
        if category == "spot":
            cost = 1
        else:
            cost = 2
        await self._limiter("trade").limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._order_response_decoder.decode(raw)

    async def post_v5_order_cancel(
        self, category: str, symbol: str, **kwargs
    ) -> BybitOrderResponse:
        """
        https://bybit-exchange.github.io/docs/v5/order/cancel-order
        """
        endpoint = "/v5/order/cancel"
        payload = {
            "category": category,
            "symbol": symbol,
            **kwargs,
        }
        if category == "spot":
            cost = 1
        else:
            cost = 2
        await self._limiter("trade").limit(key=endpoint, cost=cost)
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._order_response_decoder.decode(raw)

    def get_v5_position_list(self, category: str, **kwargs) -> BybitPositionResponse:
        endpoint = "/v5/position/list"
        payload = {
            "category": category,
            **kwargs,
        }
        self._limiter_sync("position").limit(key=endpoint, cost=1)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=True)
        return self._position_response_decoder.decode(raw)

    async def get_v5_order_realtime(self, category: str, **kwargs):
        """
        https://bybit-exchange.github.io/docs/v5/order/open-order
        """
        endpoint = "/v5/order/realtime"
        payload = {
            "category": category,
            **kwargs,
        }
        raw = await self._fetch("GET", self._base_url, endpoint, payload, signed=True)
        return self._open_orders_response_decoder.decode(raw)

    async def get_v5_order_history(self, category: str, **kwargs):
        """
        https://bybit-exchange.github.io/docs/v5/order/order-list
        """
        endpoint = "/v5/order/history"
        payload = {
            "category": category,
            **kwargs,
        }
        raw = await self._fetch("GET", self._base_url, endpoint, payload, signed=True)
        return self._order_history_response_decoder.decode(raw)

    def get_v5_account_wallet_balance(
        self, account_type: str, **kwargs
    ) -> BybitWalletBalanceResponse:
        endpoint = "/v5/account/wallet-balance"
        payload = {
            "accountType": account_type,
            **kwargs,
        }
        self._limiter_sync("account").limit(key=endpoint, cost=1)
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=True)
        return self._wallet_balance_response_decoder.decode(raw)

    async def post_v5_order_amend(
        self,
        category: str,
        symbol: str,
        orderId: str | None = None,
        orderLinkId: str | None = None,
        orderIv: str | None = None,
        triggerPrice: str | None = None,
        qty: str | None = None,
        price: str | None = None,
        tpslMode: str | None = None,
        takeProfit: str | None = None,
        stopLoss: str | None = None,
        tpTriggerBy: str | None = None,
        slTriggerBy: str | None = None,
        triggerBy: str | None = None,
        tpLimitPrice: str | None = None,
        slLimitPrice: str | None = None,
    ):
        """
        https://bybit-exchange.github.io/docs/v5/order/amend-order
        """
        endpoint = "/v5/order/amend"
        payload = {
            "category": category,
            "symbol": symbol,
            "orderId": orderId,
            "orderLinkId": orderLinkId,
            "orderIv": orderIv,
            "triggerPrice": triggerPrice,
            "qty": qty,
            "price": price,
            "tpslMode": tpslMode,
            "takeProfit": takeProfit,
            "stopLoss": stopLoss,
            "tpTriggerBy": tpTriggerBy,
            "slTriggerBy": slTriggerBy,
            "triggerBy": triggerBy,
            "tpLimitPrice": tpLimitPrice,
            "slLimitPrice": slLimitPrice,
        }
        await self._limiter("trade").limit(key=endpoint, cost=2)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._order_response_decoder.decode(raw)

    async def post_v5_order_cancel_all(
        self,
        category: str,
        symbol: str | None = None,
        baseCoin: str | None = None,
        settleCoin: str | None = None,
        orderFilter: str | None = None,
        stopOrderType: str | None = None,
    ):
        endpoint = "/v5/order/cancel-all"
        payload = {
            "category": category,
            "symbol": symbol,
            "baseCoin": baseCoin,
            "settleCoin": settleCoin,
            "orderFilter": orderFilter,
            "stopOrderType": stopOrderType,
        }
        if category == "spot":
            cost = 1
        else:
            cost = 2
        await self._limiter("trade").limit(key=endpoint, cost=cost)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._msg_decoder.decode(raw)

    def get_v5_market_kline(
        self,
        category: str,
        symbol: str,
        interval: str,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ):
        endpoint = "/v5/market/kline"
        payload = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }
        self._limiter_sync("public").limit(key=endpoint, cost=1)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=False)
        return self._kline_response_decoder.decode(raw)

    def get_v5_market_index_price_kline(
        self,
        category: str,
        symbol: str,
        interval: str,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ):
        endpoint = "/v5/market/index-price-kline"
        payload = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }
        self._limiter_sync("public").limit(key=endpoint, cost=1)
        payload = {k: v for k, v in payload.items() if v is not None}
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=False)
        return self._index_kline_response_decoder.decode(raw)

    async def post_v5_order_create_batch(
        self,
        category: str,
        request: List[Dict[str, Any]],
    ) -> BybitBatchOrderResponse:
        """
        Place multiple orders in a single request
        https://bybit-exchange.github.io/docs/v5/order/batch-place

        Maximum orders per request:
        - Option: 20 orders
        - Inverse: 20 orders
        - Linear: 20 orders
        - Spot: 10 orders
        """
        endpoint = "/v5/order/create-batch"
        payload = {
            "category": category,
            "request": request,
        }

        # Rate limit cost based on category
        if category == "spot":
            cost = len(request)  # 1 for each order in spot category
        else:
            cost = len(request) * 2  # 2 for each order in non-spot categories
        await self._limiter("trade").limit(key=endpoint, cost=cost)

        raw = await self._fetch("POST", self._base_url, endpoint, payload, signed=True)
        return self._batch_order_response_decoder.decode(raw)

    def get_v5_market_tickers(
        self,
        category: str,
        symbol: str | None = None,
        base_coin: str | None = None,
        exp_date: str | None = None,
    ) -> BybitTickersResponse:
        """
        GET /v5/market/tickers

        Query for the latest price snapshot, best bid/ask price, and trading volume
        in the last 24 hours.

        Covers: Spot / USDT contract / USDC contract / Inverse contract / Option

        Args:
            category: Product type (spot, linear, inverse, option)
            symbol: Symbol name (optional), like BTCUSDT, uppercase only
            base_coin: Base coin (optional), uppercase only. Apply to option only
            exp_date: Expiry date (optional), e.g., 25DEC22. Apply to option only

        Returns:
            BybitTickersResponse: Response containing ticker data
        """
        endpoint = "/v5/market/tickers"
        payload = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "expDate": exp_date,
        }
        payload = {k: v for k, v in payload.items() if v is not None}

        self._limiter_sync("public").limit(key=endpoint, cost=1)
        raw = self._fetch_sync("GET", self._base_url, endpoint, payload, signed=False)
        return self._tickers_response_decoder.decode(raw)



from nexustrader.exchange.bitget.exchange import BitgetExchangeManager
from nexustrader.exchange.bitget.connector import (
    BitgetPublicConnector,
    BitgetPrivateConnector,
)
from nexustrader.exchange.bitget.constants import BitgetAccountType
from nexustrader.exchange.bitget.ems import BitgetExecutionManagementSystem
from nexustrader.exchange.bitget.oms import BitgetOrderManagementSystem
from nexustrader.exchange.bitget.factory import BitgetFactory

# Auto-register factory on import
try:
    from nexustrader.exchange.registry import register_factory

    register_factory(BitgetFactory())
except ImportError:
    # Registry not available yet during bootstrap
    pass

__all__ = [
    "BitgetExchangeManager",
    "BitgetPublicConnector",
    "BitgetPrivateConnector",
    "BitgetAccountType",
    "BitgetExecutionManagementSystem",
    "BitgetOrderManagementSystem",
    "BitgetFactory",
]



import asyncio
from typing import Dict, List
from decimal import Decimal
from typing import Literal
from decimal import ROUND_HALF_UP, ROUND_CEILING, ROUND_FLOOR

from nexustrader.constants import AccountType, SubmitType
from nexustrader.schema import (
    OrderSubmit,
    InstrumentId,
    CancelAllOrderSubmit,
    CancelOrderSubmit,
)
from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import MessageBus, LiveClock
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.exchange.bitget import BitgetAccountType
from nexustrader.exchange.bitget.schema import BitgetMarket
from nexustrader.base import ExecutionManagementSystem


class BitgetExecutionManagementSystem(ExecutionManagementSystem):
    _market: Dict[str, BitgetMarket]

    def __init__(
        self,
        market: Dict[str, BitgetMarket],
        cache: AsyncCache,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        registry: OrderRegistry,
        is_mock: bool = False,
    ):
        super().__init__(
            market=market,
            cache=cache,
            msgbus=msgbus,
            clock=clock,
            task_manager=task_manager,
            registry=registry,
            is_mock=is_mock,
        )

        self._bitget_futures_account_type: BitgetAccountType = None
        self._bitget_spot_account_type: BitgetAccountType = None
        self._bitget_uta_account_type: BitgetAccountType = None

    def _build_order_submit_queues(self):
        for account_type in self._private_connectors.keys():
            if isinstance(account_type, BitgetAccountType):
                self._order_submit_queues[account_type] = asyncio.Queue()

    def _set_account_type(self):
        account_types = self._private_connectors.keys()

        if (
            BitgetAccountType.UTA_DEMO in account_types
            or BitgetAccountType.UTA in account_types
        ):
            self._bitget_uta_account_type = (
                BitgetAccountType.UTA_DEMO
                if BitgetAccountType.UTA_DEMO in account_types
                else BitgetAccountType.UTA
            )
            return

        self._bitget_futures_account_type = (
            BitgetAccountType.FUTURE_DEMO
            if BitgetAccountType.FUTURE_DEMO in account_types
            else BitgetAccountType.FUTURE
        )

        self._bitget_spot_account_type = (
            BitgetAccountType.SPOT_DEMO
            if BitgetAccountType.SPOT_DEMO in account_types
            else BitgetAccountType.SPOT
        )

    def _instrument_id_to_account_type(
        self, instrument_id: InstrumentId
    ) -> AccountType:
        if self._is_mock:
            if instrument_id.is_spot:
                return BitgetAccountType.SPOT_MOCK
            elif instrument_id.is_linear:
                return BitgetAccountType.LINEAR_MOCK
            elif instrument_id.is_inverse:
                return BitgetAccountType.INVERSE_MOCK
        else:
            if self._bitget_uta_account_type:
                return self._bitget_uta_account_type
            if instrument_id.is_spot:
                return self._bitget_spot_account_type
            elif instrument_id.is_linear or instrument_id.is_inverse:
                return self._bitget_futures_account_type

    def _submit_order(
        self,
        order: OrderSubmit | List[OrderSubmit],
        submit_type: SubmitType,
        account_type: AccountType | None = None,
    ):
        if isinstance(order, list):
            if not account_type:
                account_type = self._instrument_id_to_account_type(
                    order[0].instrument_id
                )

            # Split batch orders into chunks of 20
            for i in range(0, len(order), 20):
                batch = order[i : i + 20]
                self._order_submit_queues[account_type].put_nowait((batch, submit_type))
        else:
            if not account_type:
                account_type = self._instrument_id_to_account_type(order.instrument_id)
            self._order_submit_queues[account_type].put_nowait((order, submit_type))

    def _get_min_order_amount(self, symbol: str, market: BitgetMarket) -> Decimal:
        book = self._cache.bookl1(symbol)
        if market.spot:
            min_order_cost = float(market.info.minTradeUSDT)
            min_order_amount = min_order_cost * 1.01 / book.mid
        else:
            min_order_amt = float(market.info.minTradeNum)
            min_order_cost = float(market.info.minTradeUSDT)
            min_order_amount = max(min_order_cost * 1.02 / book.mid, min_order_amt)
        min_order_amount = self._amount_to_precision(
            symbol, min_order_amount, mode="ceil"
        )
        return min_order_amount

    def _amount_to_precision(
        self,
        symbol: str,
        amount: float,
        mode: Literal["round", "ceil", "floor"] = "round",
    ) -> Decimal:
        market = self._market[symbol]
        if market.spot:
            return super()._amount_to_precision(symbol, amount, mode)
        else:
            amount: Decimal = Decimal(str(amount))
            amount_multiplier = Decimal(market.info.sizeMultiplier)
            multiplier_count = amount / amount_multiplier

            if mode == "round":
                amount = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_HALF_UP)
                ) * amount_multiplier
            elif mode == "ceil":
                amount = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_CEILING)
                ) * amount_multiplier
            elif mode == "floor":
                amount = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_FLOOR)
                ) * amount_multiplier
            return amount

    def _price_to_precision(
        self,
        symbol: str,
        price: float,
        mode: Literal["round", "ceil", "floor"] = "round",
    ) -> Decimal:
        market = self._market[symbol]
        if market.spot:
            return super()._price_to_precision(symbol, price, mode)
        else:
            price: Decimal = Decimal(str(price))
            price_multiplier = Decimal(market.info.priceEndStep)
            multiplier_count = price / price_multiplier

            if mode == "round":
                price = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_HALF_UP)
                ) * price_multiplier
            elif mode == "ceil":
                price = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_CEILING)
                ) * price_multiplier
            elif mode == "floor":
                price = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_FLOOR)
                ) * price_multiplier
            return price

    async def _cancel_all_orders(
        self, order_submit: CancelAllOrderSubmit, account_type: AccountType
    ):
        # override the base method
        if account_type.is_uta:
            await super()._cancel_all_orders(order_submit, account_type)
        else:
            symbol = order_submit.symbol
            oids = self._cache.get_open_orders(symbol)
            for oid in oids:
                order_submit = CancelOrderSubmit(
                    symbol=symbol,
                    instrument_id=InstrumentId.from_str(symbol),
                    oid=oid,
                )
            await self._cancel_order_ws(order_submit, account_type)



import msgspec
from typing import Dict, List
from collections import defaultdict
from nexustrader.base import PublicConnector, PrivateConnector
from nexustrader.core.nautilius_core import MessageBus, LiveClock
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.core.cache import AsyncCache
from nexustrader.schema import (
    BookL1,
    Trade,
    Kline,
    BookL2,
    BookOrderData,
    FundingRate,
    IndexPrice,
    MarkPrice,
    KlineList,
    Ticker,
)
from nexustrader.constants import (
    KlineInterval,
    BookLevel,
)
from nexustrader.exchange.bybit.schema import (
    BybitKlineResponse,
    BybitKlineResponseArray,
    BybitWsMessageGeneral,
    BybitWsOrderbookDepthMsg,
    BybitOrderBook,
    BybitMarket,
    BybitWsTradeMsg,
    BybitWsTickerMsg,
    BybitWsKlineMsg,
    BybitTicker,
    BybitIndexKlineResponse,
    BybitIndexKlineResponseArray,
)
from nexustrader.exchange.bybit.rest_api import BybitApiClient
from nexustrader.exchange.bybit.websockets import BybitWSClient
from nexustrader.exchange.bybit.oms import BybitOrderManagementSystem
from nexustrader.exchange.bybit.constants import (
    BybitAccountType,
    BybitEnumParser,
)
from nexustrader.exchange.bybit.exchange import BybitExchangeManager


class BybitPublicConnector(PublicConnector):
    _api_client: BybitApiClient
    _ws_client: BybitWSClient
    _account_type: BybitAccountType

    def __init__(
        self,
        account_type: BybitAccountType,
        exchange: BybitExchangeManager,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        custom_url: str | None = None,
        enable_rate_limit: bool = True,
    ):
        if account_type in {BybitAccountType.UNIFIED, BybitAccountType.UNIFIED_TESTNET}:
            raise ValueError(
                "Please not using `BybitAccountType.UNIFIED` or `BybitAccountType.UNIFIED_TESTNET` in `PublicConnector`"
            )

        super().__init__(
            account_type=account_type,
            market=exchange.market,
            market_id=exchange.market_id,
            exchange_id=exchange.exchange_id,
            ws_client=BybitWSClient(
                account_type=account_type,
                handler=self._ws_msg_handler,
                clock=clock,
                task_manager=task_manager,
                custom_url=custom_url,
            ),
            clock=clock,
            msgbus=msgbus,
            api_client=BybitApiClient(
                clock=clock,
                testnet=account_type.is_testnet,
                enable_rate_limit=enable_rate_limit,
            ),
            task_manager=task_manager,
        )
        self._ws_msg_trade_decoder = msgspec.json.Decoder(BybitWsTradeMsg)
        self._ws_msg_orderbook_decoder = msgspec.json.Decoder(BybitWsOrderbookDepthMsg)
        self._ws_msg_general_decoder = msgspec.json.Decoder(BybitWsMessageGeneral)
        self._ws_msg_kline_decoder = msgspec.json.Decoder(BybitWsKlineMsg)
        self._ws_msg_ticker_decoder = msgspec.json.Decoder(BybitWsTickerMsg)
        self._orderbook = defaultdict(BybitOrderBook)
        self._ticker: Dict[str, BybitTicker] = defaultdict(BybitTicker)

    @property
    def market_type(self):
        if self._account_type.is_spot:
            return "_spot"
        elif self._account_type.is_linear:
            return "_linear"
        elif self._account_type.is_inverse:
            return "_inverse"
        else:
            raise ValueError(f"Unsupported BybitAccountType.{self._account_type.value}")

    def _get_category(self, market: BybitMarket):
        if market.spot:
            return "spot"
        elif market.linear:
            return "linear"
        elif market.inverse:
            return "inverse"
        else:
            raise ValueError(f"Unsupported market type: {market.type}")

    def _ws_msg_handler(self, raw: bytes):
        try:
            ws_msg: BybitWsMessageGeneral = self._ws_msg_general_decoder.decode(raw)
            if ws_msg.ret_msg == "pong":
                self._ws_client._transport.notify_user_specific_pong_received()
                self._log.debug(f"Pong received {str(ws_msg)}")
                return
            if ws_msg.success is False:
                self._log.error(f"WebSocket error: {ws_msg}")
                return

            if "orderbook.1" in ws_msg.topic:
                self._handle_orderbook(raw)
            elif "orderbook.50" in ws_msg.topic:
                self._handle_orderbook_50(raw)
            elif "publicTrade" in ws_msg.topic:
                self._handle_trade(raw)
            elif "kline" in ws_msg.topic:
                self._handle_kline(raw)
            elif "tickers" in ws_msg.topic:
                self._handle_ticker(raw)
        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding message: {str(raw)} {e}")

    def _handle_ticker(self, raw: bytes):
        msg: BybitWsTickerMsg = self._ws_msg_ticker_decoder.decode(raw)
        id = msg.data.symbol + self.market_type
        symbol = self._market_id[id]

        ticker = self._ticker[symbol]
        ticker.parse_ticker(msg)

        funding_rate = FundingRate(
            exchange=self._exchange_id,
            symbol=symbol,
            rate=float(ticker.fundingRate),
            timestamp=msg.ts,
            next_funding_time=int(ticker.nextFundingTime),
        )

        index_price = IndexPrice(
            exchange=self._exchange_id,
            symbol=symbol,
            price=ticker.indexPrice,
            timestamp=msg.ts,
        )

        mark_price = MarkPrice(
            exchange=self._exchange_id,
            symbol=symbol,
            price=ticker.markPrice,
            timestamp=msg.ts,
        )

        self._msgbus.publish(topic="funding_rate", msg=funding_rate)
        self._msgbus.publish(topic="index_price", msg=index_price)
        self._msgbus.publish(topic="mark_price", msg=mark_price)

    def _handle_kline(self, raw: bytes):
        msg: BybitWsKlineMsg = self._ws_msg_kline_decoder.decode(raw)
        id = msg.topic.split(".")[-1] + self.market_type
        symbol = self._market_id[id]
        for d in msg.data:
            interval = BybitEnumParser.parse_kline_interval(d.interval)
            kline = Kline(
                exchange=self._exchange_id,
                symbol=symbol,
                interval=interval,
                open=float(d.open),
                high=float(d.high),
                low=float(d.low),
                close=float(d.close),
                volume=float(d.volume),
                start=d.start,
                confirm=d.confirm,
                timestamp=msg.ts,
            )
            self._msgbus.publish(topic="kline", msg=kline)

    def _handle_trade(self, raw: bytes):
        msg: BybitWsTradeMsg = self._ws_msg_trade_decoder.decode(raw)
        for d in msg.data:
            id = d.s + self.market_type
            symbol = self._market_id[id]
            trade = Trade(
                exchange=self._exchange_id,
                symbol=symbol,
                price=float(d.p),
                size=float(d.v),
                timestamp=msg.ts,
                side=BybitEnumParser.parse_order_side(d.S),
            )
            self._msgbus.publish(topic="trade", msg=trade)

    def _handle_orderbook(self, raw: bytes):
        msg: BybitWsOrderbookDepthMsg = self._ws_msg_orderbook_decoder.decode(raw)
        id = msg.data.s + self.market_type
        symbol = self._market_id[id]
        res = self._orderbook[symbol].parse_orderbook_depth(msg, levels=1)

        bid, bid_size = (
            (res["bids"][0].price, res["bids"][0].size) if res["bids"] else (0, 0)
        )
        ask, ask_size = (
            (res["asks"][0].price, res["asks"][0].size) if res["asks"] else (0, 0)
        )

        bookl1 = BookL1(
            exchange=self._exchange_id,
            symbol=symbol,
            timestamp=msg.ts,
            bid=bid,
            bid_size=bid_size,
            ask=ask,
            ask_size=ask_size,
        )
        self._msgbus.publish(topic="bookl1", msg=bookl1)

    def _handle_orderbook_50(self, raw: bytes):
        msg: BybitWsOrderbookDepthMsg = self._ws_msg_orderbook_decoder.decode(raw)
        id = msg.data.s + self.market_type
        symbol = self._market_id[id]
        res = self._orderbook[symbol].parse_orderbook_depth(msg, levels=50)

        bids = res["bids"] if res["bids"] else [BookOrderData(price=0, size=0)]
        asks = res["asks"] if res["asks"] else [BookOrderData(price=0, size=0)]

        bookl2 = BookL2(
            exchange=self._exchange_id,
            symbol=symbol,
            timestamp=msg.ts,
            bids=bids,
            asks=asks,
        )
        self._msgbus.publish(topic="bookl2", msg=bookl2)

    def request_ticker(
        self,
        symbol: str,
    ) -> Ticker:
        """Request 24hr ticker data"""
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        category = self._get_category(market)
        id = market.id
        ticker_response = self._api_client.get_v5_market_tickers(
            category=category, symbol=id
        )
        for ticker in ticker_response.result.list:
            return Ticker(
                exchange=self._exchange_id,
                symbol=symbol,
                last_price=float(ticker.lastPrice),
                timestamp=ticker_response.time,
                volume=float(ticker.volume24h),
                volumeCcy=float(ticker.turnover24h),
            )

    def request_all_tickers(
        self,
    ) -> Dict[str, Ticker]:
        """Request 24hr ticker data for multiple symbols"""
        if self._account_type.is_spot:
            category = "spot"
        elif self._account_type.is_linear:
            category = "linear"
        elif self._account_type.is_inverse:
            category = "inverse"
        ticker_response = self._api_client.get_v5_market_tickers(
            category=category,
        )
        tickers = {}
        for ticker in ticker_response.result.list:
            id = ticker.symbol + self.market_type
            symbol = self._market_id.get(id)
            if not symbol:
                continue
            tickers[symbol] = Ticker(
                exchange=self._exchange_id,
                symbol=symbol,
                last_price=float(ticker.lastPrice),
                timestamp=ticker_response.time,
                volume=float(ticker.volume24h),
                volumeCcy=float(ticker.turnover24h),
            )
        return tickers

    def request_index_klines(
        self,
        symbol: str,
        interval: KlineInterval,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> KlineList:
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        if market.spot:
            raise ValueError("Spot market is not supported for index klines")
        category = self._get_category(market)
        id = market.id
        bybit_interval = BybitEnumParser.to_bybit_kline_interval(interval)
        all_klines: list[Kline] = []
        seen_timestamps: set[int] = set()
        prev_start_time: int | None = None

        while True:
            # Check for infinite loop condition
            if prev_start_time is not None and prev_start_time == start_time:
                break
            prev_start_time = start_time

            klines_response: BybitIndexKlineResponse = (
                self._api_client.get_v5_market_index_price_kline(
                    category=category,
                    symbol=id,
                    interval=bybit_interval.value,
                    limit=1000,
                    start=start_time,
                    end=end_time,
                )
            )

            # Sort klines by start time and filter out duplicates
            response_klines = sorted(
                klines_response.result.list, key=lambda k: int(k.startTime)
            )
            klines: list[Kline] = [
                self._handle_index_candlesticks(
                    symbol=symbol,
                    interval=interval,
                    kline=kline,
                    timestamp=klines_response.time,
                )
                for kline in response_klines
                if int(kline.startTime) not in seen_timestamps
            ]

            all_klines.extend(klines)
            seen_timestamps.update(int(kline.startTime) for kline in response_klines)

            # If no new klines were found, break
            if not klines:
                break

            # Update the start_time to fetch the next set of bars
            start_time = int(response_klines[-1].startTime) + 1

            # No more bars to fetch if we've reached the end time
            if end_time is not None and start_time >= end_time:
                break

        # If limit is specified, return the last 'limit' number of klines
        if limit is not None and len(all_klines) > limit:
            all_klines = all_klines[-limit:]

        kline_list = KlineList(
            all_klines,
            fields=[
                "timestamp",
                "symbol",
                "open",
                "high",
                "low",
                "close",
                "confirm",
            ],
        )
        return kline_list

    def request_klines(
        self,
        symbol: str,
        interval: KlineInterval,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> KlineList:
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        category = self._get_category(market)
        id = market.id
        bybit_interval = BybitEnumParser.to_bybit_kline_interval(interval)
        all_klines: list[Kline] = []
        seen_timestamps: set[int] = set()
        prev_start_time: int | None = None

        while True:
            # Check for infinite loop condition
            if prev_start_time is not None and prev_start_time == start_time:
                break
            prev_start_time = start_time

            klines_response: BybitKlineResponse = self._api_client.get_v5_market_kline(
                category=category,
                symbol=id,
                interval=bybit_interval.value,
                limit=1000,
                start=start_time,
                end=end_time,
            )

            # Sort klines by start time and filter out duplicates
            response_klines = sorted(
                klines_response.result.list, key=lambda k: int(k.startTime)
            )
            klines: list[Kline] = [
                self._handle_candlesticks(
                    symbol=symbol,
                    interval=interval,
                    kline=kline,
                    timestamp=klines_response.time,
                )
                for kline in response_klines
                if int(kline.startTime) not in seen_timestamps
            ]

            all_klines.extend(klines)
            seen_timestamps.update(int(kline.startTime) for kline in response_klines)

            # If no new klines were found, break
            if not klines:
                break

            # Update the start_time to fetch the next set of bars
            start_time = int(response_klines[-1].startTime) + 1

            # No more bars to fetch if we've reached the end time
            if end_time is not None and start_time >= end_time:
                break

        # If limit is specified, return the last 'limit' number of klines
        if limit is not None and len(all_klines) > limit:
            all_klines = all_klines[-limit:]

        kline_list = KlineList(
            all_klines,
            fields=[
                "timestamp",
                "symbol",
                "open",
                "high",
                "low",
                "close",
                "volume",
                "turnover",
                "confirm",
            ],
        )
        return kline_list

    async def subscribe_funding_rate(self, symbol: str):
        symbols = []
        if isinstance(symbol, str):
            symbol = [symbol]

        for s in symbol:
            market = self._market.get(s)
            if not market:
                raise ValueError(f"Symbol {s} formated wrongly, or not supported")
            symbols.append(market.id)

        await self._ws_client.subscribe_ticker(symbols)

    async def subscribe_index_price(self, symbol: str):
        symbols = []
        if isinstance(symbol, str):
            symbol = [symbol]

        for s in symbol:
            market = self._market.get(s)
            if not market:
                raise ValueError(f"Symbol {s} formated wrongly, or not supported")
            symbols.append(market.id)

        await self._ws_client.subscribe_ticker(symbols)

    async def subscribe_mark_price(self, symbol: str):
        symbols = []
        if isinstance(symbol, str):
            symbol = [symbol]

        for s in symbol:
            market = self._market.get(s)
            if not market:
                raise ValueError(f"Symbol {s} formated wrongly, or not supported")
            symbols.append(market.id)

        await self._ws_client.subscribe_ticker(symbols)

    async def subscribe_bookl1(self, symbol: str | List[str]):
        symbols = []
        if isinstance(symbol, str):
            symbol = [symbol]

        for s in symbol:
            market = self._market.get(s)
            if not market:
                raise ValueError(f"Symbol {s} formated wrongly, or not supported")
            symbols.append(market.id)

        await self._ws_client.subscribe_order_book(symbols, depth=1)

    async def subscribe_trade(self, symbol: str | List[str]):
        symbols = []
        if isinstance(symbol, str):
            symbol = [symbol]

        for s in symbol:
            market = self._market.get(s)
            if not market:
                raise ValueError(f"Symbol {s} formated wrongly, or not supported")
            symbols.append(market.id)

        await self._ws_client.subscribe_trade(symbols)

    async def subscribe_kline(self, symbol: str | List[str], interval: KlineInterval):
        symbols = []
        if isinstance(symbol, str):
            symbol = [symbol]

        for s in symbol:
            market = self._market.get(s)
            if not market:
                raise ValueError(f"Symbol {s} formated wrongly, or not supported")
            symbols.append(market.id)

        interval = BybitEnumParser.to_bybit_kline_interval(interval)
        await self._ws_client.subscribe_kline(symbols, interval)

    async def subscribe_bookl2(self, symbol: str | List[str], level: BookLevel):
        if level != BookLevel.L50:
            raise ValueError(f"Unsupported book level: {level}")

        symbols = []
        if isinstance(symbol, str):
            symbol = [symbol]

        for s in symbol:
            market = self._market.get(s)
            if not market:
                raise ValueError(f"Symbol {s} formated wrongly, or not supported")
            symbols.append(market.id)

        await self._ws_client.subscribe_order_book(symbols, depth=50)

    def _handle_index_candlesticks(
        self,
        symbol: str,
        interval: KlineInterval,
        kline: BybitIndexKlineResponseArray,
        timestamp: int,
    ) -> Kline:
        local_timestamp = self._clock.timestamp_ms()
        confirm = (
            True
            if local_timestamp >= int(kline.startTime) + interval.seconds * 1000 - 1
            else False
        )
        return Kline(
            exchange=self._exchange_id,
            symbol=symbol,
            interval=interval,
            open=float(kline.openPrice),
            high=float(kline.highPrice),
            low=float(kline.lowPrice),
            close=float(kline.closePrice),
            start=int(kline.startTime),
            timestamp=timestamp,
            confirm=confirm,
        )

    def _handle_candlesticks(
        self,
        symbol: str,
        interval: KlineInterval,
        kline: BybitKlineResponseArray,
        timestamp: int,
    ) -> Kline:
        local_timestamp = self._clock.timestamp_ms()
        confirm = (
            True
            if local_timestamp >= int(kline.startTime) + interval.seconds * 1000 - 1
            else False
        )
        return Kline(
            exchange=self._exchange_id,
            symbol=symbol,
            interval=interval,
            open=float(kline.openPrice),
            high=float(kline.highPrice),
            low=float(kline.lowPrice),
            close=float(kline.closePrice),
            volume=float(kline.volume),
            start=int(kline.startTime),
            turnover=float(kline.turnover),
            timestamp=timestamp,
            confirm=confirm,
        )


class BybitPrivateConnector(PrivateConnector):
    _account_type: BybitAccountType
    _market: Dict[str, BybitMarket]
    _market_id: Dict[str, str]
    _api_client: BybitApiClient
    _oms: BybitOrderManagementSystem

    def __init__(
        self,
        account_type: BybitAccountType,
        exchange: BybitExchangeManager,
        cache: AsyncCache,
        registry: OrderRegistry,
        clock: LiveClock,
        msgbus: MessageBus,
        task_manager: TaskManager,
        enable_rate_limit: bool = True,
        **kwargs,
    ):
        if not exchange.api_key or not exchange.secret:
            raise ValueError("API key and secret are required for private endpoints")

        if account_type not in {
            BybitAccountType.UNIFIED,
            BybitAccountType.UNIFIED_TESTNET,
        }:
            raise ValueError(
                "Please using `BybitAccountType.UNIFIED` or `BybitAccountType.UNIFIED_TESTNET` in `PrivateConnector`"
            )

        api_client = BybitApiClient(
            clock=clock,
            api_key=exchange.api_key,
            secret=exchange.secret,
            testnet=account_type.is_testnet,
            enable_rate_limit=enable_rate_limit,
            **kwargs,
        )

        oms = BybitOrderManagementSystem(
            account_type=account_type,
            api_key=exchange.api_key,
            secret=exchange.secret,
            market=exchange.market,
            market_id=exchange.market_id,
            registry=registry,
            cache=cache,
            api_client=api_client,
            exchange_id=exchange.exchange_id,
            clock=clock,
            msgbus=msgbus,
            task_manager=task_manager,
            enable_rate_limit=enable_rate_limit,
        )

        super().__init__(
            account_type=account_type,
            market=exchange.market,
            api_client=api_client,
            task_manager=task_manager,
            oms=oms,
        )

    async def connect(self):
        await self._oms._ws_api_client.connect()
        await self._oms._ws_client.subscribe_order()
        await self._oms._ws_client.subscribe_position()
        await self._oms._ws_client.subscribe_wallet()



from abc import ABC
from typing import Optional
from curl_cffi import requests
from nexustrader.core.nautilius_core import LiveClock, Logger
from nexustrader.constants import RateLimiter, RateLimiterSync
from nexustrader.base.retry import RetryManager


class ApiClient(ABC):
    def __init__(
        self,
        clock: LiveClock,
        api_key: str = None,
        secret: str = None,
        timeout: int = 10,
        rate_limiter: RateLimiter = None,
        rate_limiter_sync: RateLimiterSync = None,
        retry_manager: RetryManager = None,
    ):
        self._api_key = api_key
        self._secret = secret
        self._timeout = timeout
        self._log = Logger(name=type(self).__name__)
        self._session: Optional[requests.AsyncSession] = None
        self._sync_session: Optional[requests.Session] = None
        self._clock = clock
        self._limiter = rate_limiter
        self._limiter_sync = rate_limiter_sync
        self._retry_manager: RetryManager = retry_manager

    def _init_session(self, base_url: str | None = None):
        if self._session is None:
            self._session = requests.AsyncSession(
                base_url=base_url if base_url else "", timeout=self._timeout
            )

    def _get_rate_limit_cost(self, cost: int = 1):
        return cost

    def _init_sync_session(self, base_url: str | None = None):
        if self._sync_session is None:
            self._sync_session = requests.Session(
                base_url=base_url if base_url else "", timeout=self._timeout
            )

    async def close_session(self):
        """Close the session"""
        if self._session:
            await self._session.close()
            self._session = None
        if self._sync_session:
            self._sync_session.close()
            self._sync_session = None



import hmac
import base64
import asyncio

from typing import Any, Callable, List, Dict

from nexustrader.base import WSClient
from nexustrader.core.entity import TaskManager
from nexustrader.core.nautilius_core import LiveClock
from nexustrader.exchange.bitget.constants import (
    BitgetAccountType,
    BitgetKlineInterval,
    BitgetRateLimiter,
)


class BitgetWSClient(WSClient):
    def __init__(
        self,
        account_type: BitgetAccountType,
        handler: Callable[..., Any],
        task_manager: TaskManager,
        clock: LiveClock,
        api_key: str | None = None,
        secret: str | None = None,
        passphrase: str | None = None,
        custom_url: str | None = None,
    ):
        self._api_key = api_key
        self._secret = secret
        self._passphrase = passphrase
        self._account_type = account_type
        self._authed = False

        if custom_url:
            url = custom_url
        elif self.is_private:
            url = f"{account_type.stream_url}/private"
        else:
            url = f"{account_type.stream_url}/public"

        super().__init__(
            url,
            handler=handler,
            task_manager=task_manager,
            clock=clock,
            specific_ping_msg=b"ping",
            auto_ping_strategy="ping_periodically",
            ping_idle_timeout=30,
            ping_reply_timeout=5,
        )

    @property
    def is_private(self):
        return (
            self._api_key is not None
            or self._secret is not None
            or self._passphrase is not None
        )

    def _get_auth_payload(self):
        timestamp = self._clock.timestamp_ms()
        message = f"{timestamp}GET/user/verify"

        # Create HMAC-SHA256 signature
        mac = hmac.new(
            bytes(self._secret, encoding="utf8"),
            bytes(message, encoding="utf-8"),
            digestmod="sha256",
        )

        # Get the digest and encode with base64
        signature = base64.b64encode(mac.digest()).decode("utf-8")

        # Form the login payload
        payload = {
            "op": "login",
            "args": [
                {
                    "apiKey": self._api_key,
                    "passphrase": self._passphrase,  # If required
                    "timestamp": timestamp,
                    "sign": signature,
                }
            ],
        }

        return payload

    async def _auth(self):
        if not self._authed:
            self._send(self._get_auth_payload())
            self._authed = True
            await asyncio.sleep(5)

    def _send_payload(self, params: List[str], chunk_size: int = 100):
        # Split params into chunks of 100 if length exceeds 100
        params_chunks = [
            params[i : i + chunk_size] for i in range(0, len(params), chunk_size)
        ]

        for chunk in params_chunks:
            payload = {"op": "subscribe", "args": chunk}
            self._send(payload)

    async def _subscribe(self, params: List[Dict[str, Any]], auth: bool = False):
        params = [param for param in params if param not in self._subscriptions]

        for param in params:
            self._subscriptions.append(param)
            formatted_param = ".".join(param.values())
            self._log.debug(f"Subscribing to {formatted_param}...")

        await self.connect()
        if auth:
            await self._auth()
        self._send_payload(params)

    async def subscribe_depth(self, symbols: List[str], inst_type: str, channel: str):
        if channel not in ["books1", "books5", "books15"]:
            raise ValueError(f"Invalid channel: {channel}")

        params = [
            {"instType": inst_type, "channel": channel, "instId": symbol}
            for symbol in symbols
        ]
        await self._subscribe(params)

    async def subscribe_candlesticks(
        self, symbols: List[str], inst_type: str, interval: BitgetKlineInterval
    ):
        params = [
            {"instType": inst_type, "channel": interval.value, "instId": symbol}
            for symbol in symbols
        ]
        await self._subscribe(params)

    async def subscribe_trade(self, symbols: List[str], inst_type: str):
        params = [
            {"instType": inst_type, "channel": "trade", "instId": symbol}
            for symbol in symbols
        ]
        await self._subscribe(params)

    async def subscribe_ticker(self, symbols: List[str], inst_type: str):
        params = [
            {"instType": inst_type, "channel": "ticker", "instId": symbol}
            for symbol in symbols
        ]
        await self._subscribe(params)

    async def subscribe_account(self, inst_types: List[str] | str):
        if isinstance(inst_types, str):
            inst_types = [inst_types]
        params = [
            {"instType": inst_type, "channel": "account", "coin": "default"}
            for inst_type in inst_types
        ]
        await self._subscribe(params, auth=True)

    async def subscribe_positions(self, inst_types: List[str] | str):
        if isinstance(inst_types, str):
            inst_types = [inst_types]
        params = [
            {"instType": inst_type, "channel": "positions", "instId": "default"}
            for inst_type in inst_types
        ]
        await self._subscribe(params, auth=True)

    async def subscribe_orders(self, inst_types: List[str] | str):
        if isinstance(inst_types, str):
            inst_types = [inst_types]
        params = [
            {"instType": inst_type, "channel": "orders", "instId": "default"}
            for inst_type in inst_types
        ]
        await self._subscribe(params, auth=True)

    ############ UTA subscribe ###########
    ### Better Performance with v3 API ###

    async def subscribe_depth_v3(self, symbols: List[str], inst_type: str, topic: str):
        if topic not in ["books1", "books5", "books15"]:
            raise ValueError(f"Invalid channel: {topic}")

        params = [
            {"instType": inst_type, "topic": topic, "symbol": symbol}
            for symbol in symbols
        ]
        await self._subscribe(params)

    async def subscribe_trades_v3(self, symbols: List[str], inst_type: str):
        params = [
            {"instType": inst_type, "topic": "publicTrade", "symbol": symbol}
            for symbol in symbols
        ]
        await self._subscribe(params)

    async def subscribe_candlestick_v3(
        self, symbols: List[str], inst_type: str, interval: BitgetKlineInterval
    ):
        params = [
            {
                "instType": inst_type,
                "topic": "kline",
                "symbol": symbol,
                "interval": interval.value,
            }
            for symbol in symbols
        ]
        await self._subscribe(params)

    async def _resubscribe(self):
        if self.is_private:
            self._authed = False
            await self._auth()
        self._send_payload(self._subscriptions)

    async def subscribe_v3_order(self):
        params = [{"instType": "UTA", "topic": "order"}]
        await self._subscribe(params, auth=True)

    async def subscribe_v3_position(self):
        params = [{"instType": "UTA", "topic": "position"}]
        await self._subscribe(params, auth=True)

    async def subscribe_v3_account(self):
        params = [{"instType": "UTA", "topic": "account"}]
        await self._subscribe(params, auth=True)


class BitgetWSApiClient(WSClient):
    def __init__(
        self,
        account_type: BitgetAccountType,
        api_key: str,
        secret: str,
        passphrase: str,
        handler: Callable[..., Any],
        task_manager: TaskManager,
        clock: LiveClock,
        enable_rate_limit: bool,
    ):
        self._api_key = api_key
        self._secret = secret
        self._passphrase = passphrase
        self._account_type = account_type
        self._authed = False

        # Use V2 WebSocket API for trading
        url = f"{account_type.stream_url}/private"

        self._limiter = BitgetRateLimiter(
            enable_rate_limit=enable_rate_limit,
        )

        super().__init__(
            url,
            handler=handler,
            task_manager=task_manager,
            clock=clock,
            specific_ping_msg=b"ping",
            auto_ping_strategy="ping_periodically",
            ping_idle_timeout=30,
            ping_reply_timeout=5,
        )

    def _get_auth_payload(self):
        timestamp = self._clock.timestamp_ms()
        message = f"{timestamp}GET/user/verify"

        # Create HMAC-SHA256 signature
        mac = hmac.new(
            bytes(self._secret, encoding="utf8"),
            bytes(message, encoding="utf-8"),
            digestmod="sha256",
        )

        # Get the digest and encode with base64
        signature = base64.b64encode(mac.digest()).decode("utf-8")

        # Form the login payload
        payload = {
            "op": "login",
            "args": [
                {
                    "apiKey": self._api_key,
                    "passphrase": self._passphrase,
                    "timestamp": timestamp,
                    "sign": signature,
                }
            ],
        }

        return payload

    async def _auth(self):
        if not self._authed:
            self._send(self._get_auth_payload())
            self._authed = True
            await asyncio.sleep(5)

    def _submit(
        self, id: str, instType: str, instId: str, channel: str, params: Dict[str, Any]
    ):
        payload = {
            "op": "trade",
            "args": [
                {
                    "id": id,
                    "instType": instType,
                    "instId": instId,
                    "channel": channel,
                    "params": params,
                }
            ],
        }
        self._log.debug(str(payload))
        self._send(payload)

    def _uta_submit(
        self, id: str, topic: str, category: str, args: List[Dict[str, Any]]
    ):
        payload = {
            "op": "trade",
            "id": id,
            "topic": topic,
            "category": category,
            "args": args,
        }
        self._log.debug(str(payload))
        self._send(payload)

    async def spot_place_order(
        self,
        id: str,
        instId: str,
        orderType: str,
        side: str,
        size: str,
        force: str,
        **kwargs,
    ):
        params = {
            "orderType": orderType,
            "side": side,
            "size": size,
            "force": force,
            **kwargs,
        }

        await self._limiter("/api/v2/spot/trade/place-order").limit(
            key="spot_place_order"
        )
        self._submit(
            id=id, instType="SPOT", instId=instId, channel="place-order", params=params
        )

    async def spot_cancel_order(
        self,
        id: str,
        instId: str,
        clientOid: str,
    ):
        params = {
            "clientOid": clientOid,
        }

        await self._limiter("/api/v2/spot/trade/cancel-order").limit(
            key="spot_cancel_order"
        )
        self._submit(
            id=id, instType="SPOT", instId=instId, channel="cancel-order", params=params
        )

    async def future_place_order(
        self,
        id: str,
        instId: str,
        orderType: str,
        side: str,
        size: str,
        force: str,
        marginCoin: str,
        marginMode: str,
        instType: str,
        **kwargs,
    ):
        params = {
            "orderType": orderType,
            "side": side,
            "size": size,
            "force": force,
            "marginCoin": marginCoin,
            "marginMode": marginMode,
            **kwargs,
        }

        await self._limiter("/api/v2/mix/order/place-order").limit(
            key="future_place_order"
        )
        self._submit(
            id=id,
            instType=instType,
            instId=instId,
            channel="place-order",
            params=params,
        )

    async def future_cancel_order(
        self,
        id: str,
        instId: str,
        clientOid: str,
        instType: str,
    ):
        params = {
            "clientOid": clientOid,
        }

        await self._limiter("/api/v2/mix/order/cancel-order").limit(
            key="future_cancel_order"
        )
        self._submit(
            id=id,
            instType=instType,
            instId=instId,
            channel="cancel-order",
            params=params,
        )

    async def uta_place_order(
        self,
        id: str,
        category: str,
        symbol: str,
        orderType: str,
        qty: str,
        price: str,
        side: str,
        **kwargs,
    ):
        args = [
            {
                "orderType": orderType,
                "price": price,
                "qty": qty,
                "side": side,
                "symbol": symbol,
                **kwargs,
            }
        ]

        await self._limiter("/api/v3/trade/place-order").limit(key="uta_place_order")
        self._uta_submit(id=f"n{id}", topic="place-order", category=category, args=args)

    async def uta_cancel_order(
        self,
        id: str,
        clientOid: str,
    ):
        args = [
            {
                "clientOid": clientOid,
            }
        ]

        await self._limiter("/api/v3/trade/cancel-order").limit(key="uta_cancel_order")
        self._uta_submit(
            id=f"c{id}",
            topic="cancel-order",
            category="",  # No category specified in UTA cancel order
            args=args,
        )

    async def connect(self):
        await super().connect()
        await self._auth()

    async def _resubscribe(self):
        self._authed = False
        await self._auth()


# async def main():
#     from nexustrader.constants import settings
#     from nexustrader.core.entity import TaskManager
#     from nexustrader.core.nautilius_core import setup_nautilus_core, UUID4

#     API_KEY = settings.BITGET.DEMO1.API_KEY
#     SECRET = settings.BITGET.DEMO1.SECRET
#     PASSPHRASE = settings.BITGET.DEMO1.PASSPHRASE

#     log_guard, _, clock = setup_nautilus_core(  # noqa
#         trader_id="bnc-test",
#         level_stdout="DEBUG",
#     )

#     task_manager = TaskManager(
#         loop=asyncio.get_event_loop(),
#     )

#     ws_api_client = BitgetWSApiClient(
#         account_type=BitgetAccountType.UTA_DEMO,
#         api_key=API_KEY,
#         secret=SECRET,
#         passphrase=PASSPHRASE,
#         handler=lambda msg: print(msg),
#         task_manager=task_manager,
#         clock=clock,
#         enable_rate_limit=True,
#     )

#     await ws_api_client.connect()
# await ws_api_client.spot_place_order(
#     id=UUID4().value,
#     instId="BTCUSDT",
#     orderType="limit",
#     side="buy",
#     size="0.001",
#     price="116881",
#     force="gtc",
# )

# await ws_api_client.spot_cancel_order(
#     id="9ee1e7d2-99a1-4ff5-82f7-473f02ff38e7",
#     instId="BTCUSDT",
#     orderId="1344423929063153665",
# )

# await ws_api_client.uta_place_order(
#     id=UUID4().value,
#     category="spot",  # SPOT MARGIN USDT-FUTURES COIN-FUTURES USDC-FUTURES
#     symbol="BTCUSDT",
#     orderType="limit",
#     qty="0.001",
#     price="110536",
#     side="buy",
# )

# await ws_api_client.uta_cancel_order(
#     id="3c14b801-ecc0-47f9-8126-56604d9f4c33",
#     instId="BTCUSDT",
#     orderId="1344515561982640128",
# )

# await task_manager.wait()

# {
#     "event": "trade",
#     "arg": [
#         {
#             "id": "30862609-4cb0-45cb-8562-992aaf232ee8",
#             "instType": "SPOT",
#             "channel": "place-order",
#             "instId": "BTCUSDT",
#             "params": {
#                 "orderId": "1344419561630867456",
#                 "clientOid": "0021c683-5e7e-416e-8dbb-dd34a0c3f97c",
#             },
#         }
#     ],
#     "code": 0,
#     "msg": "Success",
#     "ts": 1756260522296,
# }

# {
#     "event": "error",
#     "arg": [
#         {
#             "id": "1c6dbaae-4599-4371-9517-f800bc2f0dfb",
#             "instType": "SPOT",
#             "channel": "place-order",
#             "instId": "BTCUSDT",
#             "params": {
#                 "orderType": "limit",
#                 "side": "buy",
#                 "force": "gtc",
#                 "price": "80000",
#                 "size": "0.000001",
#             },
#         }
#     ],
#     "code": 43027,
#     "msg": "The minimum order value 1 is not met",
#     "ts": 1756260596285,
# }

# {
#     "event": "trade",
#     "arg": [
#         {
#             "id": "30862609-4cb0-45cb-8562-992aaf232ee8",
#             "instType": "SPOT",
#             "channel": "cancel-order",
#             "instId": "BTCUSDT",
#             "params": {"orderId": "1344419561630867456"},
#         }
#     ],
#     "code": 0,
#     "msg": "Success",
#     "ts": 1756260769404,
# }

# {
#     "event": "error",
#     "arg": [
#         {
#             "id": "9ee1e7d2-99a1-4ff5-82f7-473f02ff38e7",
#             "instType": "SPOT",
#             "channel": "cancel-order",
#             "instId": "BTCUSDT",
#             "params": {"orderId": "1344423929063153665"},
#         }
#     ],
#     "code": 43001,
#     "msg": "The order does not exist",
#     "ts": 1756261603661,
# }

##### UTA ACCOUNT #####

# {
#     "event": "error",
#     "id": "bb63695d-b11f-4c6d-8166-460c612dc252",
#     "code": "41101",
#     "msg": "Param category=SPOT error",
# }

# {
#     "event": "error",
#     "id": "401822f1-4014-4bf0-af07-723fcbe391d4",
#     "code": "25206",
#     "msg": "BTC/USDT trading price cannot exceed 5%",
# }

# {
#     "event": "trade",
#     "id": "90ab565b-f863-4377-b275-08020b6c0536",
#     "category": "spot",
#     "topic": "place-order",
#     "args": [
#         {
#             "symbol": "BTCUSDT",
#             "orderId": "1344507887740108800",
#             "clientOid": "1344507887740108801",
#             "cTime": "1756281580862",
#         }
#     ],
#     "code": "0",
#     "msg": "Success",
#     "ts": "1756281580865",
# }

# {
#     "event": "error",
#     "id": "90ab565b-f863-4377-b275-08020b6c0536",
#     "code": "25204",
#     "msg": "Order does not exist",
# }

# {
#     "event": "trade",
#     "id": "3c14b801-ecc0-47f9-8126-56604d9f4c33",
#     "topic": "cancel-order",
#     "args": [
#         {"orderId": "1344515561982640128", "clientOid": "1344515561982640129"}
#     ],
#     "code": "0",
#     "msg": "Success",
#     "ts": "1756283452147",
# }


# if __name__ == "__main__":
#     asyncio.run(main())



import msgspec
from typing import Dict, List
from decimal import Decimal
from typing import Literal
from decimal import ROUND_HALF_UP, ROUND_CEILING, ROUND_FLOOR
from nexustrader.error import PositionModeError
from nexustrader.core.nautilius_core import LiveClock, MessageBus
from nexustrader.core.cache import AsyncCache
from nexustrader.base import OrderManagementSystem
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.schema import (
    Order,
    Position,
    BatchOrderSubmit,
)
from nexustrader.constants import (
    ExchangeType,
    OrderSide,
    OrderStatus,
    OrderType,
    TimeInForce,
    TriggerType,
)
from nexustrader.exchange.bitget.schema import (
    BitgetMarket,
    BitgetWsGeneralMsg,
    BitgetWsUtaGeneralMsg,
    BitgetWsArgMsg,
    BitgetOrderWsMsg,
    BitgetPositionWsMsg,
    BitgetUtaOrderWsMsg,
    BitgetUtaPositionWsMsg,
    BitgetSpotAccountWsMsg,
    BitgetFuturesAccountWsMsg,
    BitgetUtaAccountWsMsg,
    BitgetWsApiGeneralMsg,
    BitgetWsApiArgMsg,
    BitgetWsApiUtaGeneralMsg,
)
from nexustrader.exchange.bitget.rest_api import BitgetApiClient
from nexustrader.exchange.bitget.websockets import BitgetWSClient, BitgetWSApiClient
from nexustrader.exchange.bitget.constants import (
    BitgetAccountType,
    BitgetInstType,
    BitgetUtaInstType,
    BitgetEnumParser,
    BitgetPositionSide,
)


class BitgetOrderManagementSystem(OrderManagementSystem):
    _inst_type_map = {
        BitgetInstType.SPOT: "spot",
        BitgetInstType.USDC_FUTURES: "linear",
        BitgetInstType.USDT_FUTURES: "linear",
        BitgetInstType.COIN_FUTURES: "inverse",
    }
    _uta_inst_type_map = {
        BitgetUtaInstType.SPOT: "spot",
        BitgetUtaInstType.USDC_FUTURES: "linear",
        BitgetUtaInstType.USDT_FUTURES: "linear",
        BitgetUtaInstType.COIN_FUTURES: "inverse",
    }
    _ws_client: BitgetWSClient
    _account_type: BitgetAccountType
    _market: Dict[str, BitgetMarket]
    _market_id: Dict[str, str]
    _api_client: BitgetApiClient

    def __init__(
        self,
        account_type: BitgetAccountType,
        api_key: str,
        secret: str,
        passphrase: str,
        market: Dict[str, BitgetMarket],
        market_id: Dict[str, str],
        registry: OrderRegistry,
        cache: AsyncCache,
        api_client: BitgetApiClient,
        exchange_id: ExchangeType,
        clock: LiveClock,
        msgbus: MessageBus,
        task_manager: TaskManager,
        max_slippage: float,
        enable_rate_limit: bool,
    ):
        super().__init__(
            account_type=account_type,
            market=market,
            market_id=market_id,
            registry=registry,
            cache=cache,
            api_client=api_client,
            ws_client=BitgetWSClient(
                account_type=account_type,
                handler=self._ws_uta_msg_handler
                if account_type.is_uta
                else self._ws_msg_handler,
                clock=clock,
                task_manager=task_manager,
                api_key=api_key,
                secret=secret,
                passphrase=passphrase,
            ),
            exchange_id=exchange_id,
            clock=clock,
            msgbus=msgbus,
        )

        self._ws_api_client = BitgetWSApiClient(
            account_type=account_type,
            api_key=api_key,
            secret=secret,
            passphrase=passphrase,
            handler=self._ws_uta_api_msg_handler
            if account_type.is_uta
            else self._ws_api_msg_handler,
            task_manager=task_manager,
            clock=clock,
            enable_rate_limit=enable_rate_limit,
        )

        self._max_slippage = max_slippage
        self._ws_msg_general_decoder = msgspec.json.Decoder(BitgetWsGeneralMsg)
        self._ws_msg_uta_general_decoder = msgspec.json.Decoder(BitgetWsUtaGeneralMsg)
        self._ws_msg_orders_decoder = msgspec.json.Decoder(BitgetOrderWsMsg)
        self._ws_msg_uta_orders_decoder = msgspec.json.Decoder(BitgetUtaOrderWsMsg)
        self._ws_msg_positions_decoder = msgspec.json.Decoder(BitgetPositionWsMsg)
        self._ws_msg_uta_positions_decoder = msgspec.json.Decoder(
            BitgetUtaPositionWsMsg
        )

        self._ws_msg_spot_account_decoder = msgspec.json.Decoder(BitgetSpotAccountWsMsg)
        self._ws_msg_futures_account_decoder = msgspec.json.Decoder(
            BitgetFuturesAccountWsMsg
        )
        self._ws_msg_uta_account_decoder = msgspec.json.Decoder(BitgetUtaAccountWsMsg)

        self._ws_api_general_decoder = msgspec.json.Decoder(BitgetWsApiGeneralMsg)
        self._ws_api_uta_general_decoder = msgspec.json.Decoder(
            BitgetWsApiUtaGeneralMsg
        )

    def _ws_uta_api_msg_handler(self, raw: bytes):
        if raw == b"pong":
            self._ws_api_client._transport.notify_user_specific_pong_received()
            self._log.debug(f"Pong received: `{raw.decode()}`")
            return
        try:
            ws_msg = self._ws_api_uta_general_decoder.decode(raw)

            if ws_msg.is_id_msg:
                self._handle_id_messages(ws_msg)
            elif ws_msg.is_error_msg:
                self._log.error(f"login Error: {ws_msg.error_msg}")

        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding message: {str(raw)} {e}")

    def _ws_api_msg_handler(self, raw: bytes):
        if raw == b"pong":
            self._ws_api_client._transport.notify_user_specific_pong_received()
            self._log.debug(f"Pong received: `{raw.decode()}`")
            return

        try:
            ws_msg = self._ws_api_general_decoder.decode(raw)

            if ws_msg.is_arg_msg:
                self._handle_arg_messages(ws_msg)
            elif ws_msg.is_error_msg:
                self._log.error(f"login Error: {ws_msg.error_msg}")

        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding message: {str(raw)} {e}")

    def _handle_id_messages(self, ws_msg: BitgetWsApiUtaGeneralMsg):
        """Handle argument messages for place and cancel orders"""
        if ws_msg.id.startswith("n"):
            self._handle_uta_place_order_response(ws_msg)
        else:
            self._handle_uta_cancel_order_response(ws_msg)

    def _handle_arg_messages(self, ws_msg: BitgetWsApiGeneralMsg):
        """Handle argument messages for place and cancel orders"""
        for arg_msg in ws_msg.arg:
            if arg_msg.is_place_order:
                self._handle_place_order_response(ws_msg, arg_msg)
            elif arg_msg.is_cancel_order:
                self._handle_cancel_order_response(ws_msg, arg_msg)

    def _handle_uta_place_order_response(self, ws_msg: BitgetWsApiUtaGeneralMsg):
        oid = ws_msg.oid
        tmp_order = self._registry.get_tmp_order(oid)
        if not tmp_order:
            return
        ts = self._clock.timestamp_ms()

        if ws_msg.is_success:
            for arg_msg in ws_msg.args:
                ordId = arg_msg.orderId
                self._log.debug(
                    f"[{tmp_order.symbol}] placing order success: oid: {oid} id: {ordId}"
                )
                order = self._create_order_from_tmp(
                    tmp_order, oid, ordId, OrderStatus.PENDING, ts
                )
                self.order_status_update(order)  # INITIALIZED -> PENDING
        else:
            self._log.error(
                f"[{tmp_order.symbol}] new order failed: oid: {oid} {ws_msg.error_msg}"
            )
            order = self._create_order_from_tmp(
                tmp_order, oid, None, OrderStatus.FAILED, ts
            )
            self.order_status_update(order)

    def _handle_uta_cancel_order_response(self, ws_msg: BitgetWsApiUtaGeneralMsg):
        """Handle cancel order response"""
        oid = ws_msg.oid
        tmp_order = self._registry.get_tmp_order(oid)
        if not tmp_order:
            return
        ts = self._clock.timestamp_ms()

        if ws_msg.is_success:
            for arg_msg in ws_msg.args:
                ordId = arg_msg.orderId
                self._log.debug(
                    f"[{tmp_order.symbol}] canceling order success: oid: {oid} id: {ordId}"
                )
                order = self._create_order_from_tmp(
                    tmp_order, oid, ordId, OrderStatus.CANCELING, ts
                )
                self.order_status_update(order)  # SOME STATUS -> CANCELING
        else:
            self._log.error(
                f"[{tmp_order.symbol}] canceling order failed: oid: {oid} {ws_msg.error_msg}"
            )
            order = self._create_order_from_tmp(
                tmp_order, oid, None, OrderStatus.CANCEL_FAILED, ts
            )
            self.order_status_update(order)

    def _handle_place_order_response(
        self, ws_msg: BitgetWsApiGeneralMsg, arg_msg: BitgetWsApiArgMsg
    ):
        """Handle place order response"""
        oid = arg_msg.id
        tmp_order = self._registry.get_tmp_order(oid)
        if not tmp_order:
            return
        ts = self._clock.timestamp_ms()

        if ws_msg.is_success:
            ordId = arg_msg.params.orderId
            self._log.debug(
                f"[{tmp_order.symbol}] placing order success: oid: {oid} id: {ordId}"
            )
            order = self._create_order_from_tmp(
                tmp_order, oid, ordId, OrderStatus.PENDING, ts
            )
            self.order_status_update(order)  # INITIALIZED -> PENDING
        else:
            self._log.error(
                f"[{tmp_order.symbol}] new order failed: oid: {oid} {ws_msg.error_msg}"
            )
            order = self._create_order_from_tmp(
                tmp_order, oid, None, OrderStatus.FAILED, ts
            )
            self.order_status_update(order)  # INITIALIZED -> FAILED

    def _handle_cancel_order_response(
        self, ws_msg: BitgetWsApiGeneralMsg, arg_msg: BitgetWsApiArgMsg
    ):
        """Handle cancel order response"""
        oid = arg_msg.id
        tmp_order = self._registry.get_tmp_order(oid)
        if not tmp_order:
            return
        ts = self._clock.timestamp_ms()

        if ws_msg.is_success:
            ordId = arg_msg.params.orderId
            self._log.debug(
                f"[{tmp_order.symbol}] canceling order success: oid: {oid} id: {ordId}"
            )
            order = self._create_order_from_tmp(
                tmp_order, oid, ordId, OrderStatus.CANCELING, ts
            )
            self.order_status_update(order)
        else:
            self._log.error(
                f"[{tmp_order.symbol}] canceling order failed: oid: {oid} {ws_msg.error_msg}"
            )
            order = self._create_order_from_tmp(
                tmp_order, oid, None, OrderStatus.CANCEL_FAILED, ts
            )
            self.order_status_update(order)

    def _create_order_from_tmp(
        self,
        tmp_order: Order,
        oid: str,
        order_id: str | None,
        status: OrderStatus,
        timestamp: int,
    ) -> Order:
        """Create Order object from temporary order data"""
        return Order(
            exchange=self._exchange_id,
            symbol=tmp_order.symbol,
            oid=oid,
            eid=order_id,
            side=tmp_order.side,
            status=status,
            amount=tmp_order.amount,
            timestamp=timestamp,
            type=tmp_order.type,
            price=tmp_order.price,
            time_in_force=tmp_order.time_in_force,
            reduce_only=tmp_order.reduce_only,
        )

    def _inst_type_suffix(self, inst_type: BitgetInstType):
        return self._inst_type_map[inst_type]

    def _uta_inst_type_suffix(self, category: BitgetUtaInstType) -> str:
        """Convert UTA category to internal suffix"""
        return self._uta_inst_type_map[category]

    def _get_inst_type(self, market: BitgetMarket):
        if market.spot:
            return "SPOT"
        elif market.linear:
            return "USDT-FUTURES" if market.quote == "USDT" else "USDC-FUTURES"
        elif market.inverse:
            return "COIN-FUTURES"

    def _init_account_balance(self):
        """Initialize account balance"""
        pass

    def _init_position(self):
        """Initialize position"""
        # Bitget has different product types for different instrument types

        if self._account_type.is_uta:
            # For UTA accounts, use the v3 position API
            categories = ["USDT-FUTURES", "USDC-FUTURES", "COIN-FUTURES"]

            for category in categories:
                response = self._api_client.get_api_v3_position_current_position(
                    category=category
                )

                if not response.data.list:
                    continue

                for pos_data in response.data.list:
                    # Skip positions with zero total
                    if float(pos_data.total) == 0:
                        continue

                    # Check position mode - only support one-way mode for UTA
                    if pos_data.holdMode != "one_way_mode":
                        raise PositionModeError(
                            f"Only one-way mode is supported for UTA accounts. Current mode: {pos_data.holdMode}"
                        )

                    # Determine suffix based on category
                    if category == "USDT-FUTURES":
                        inst_type_suffix = "linear"
                    elif category == "USDC-FUTURES":
                        inst_type_suffix = "linear"
                    elif category == "COIN-FUTURES":
                        inst_type_suffix = "inverse"

                    # Map symbol from Bitget format to our internal format
                    symbol = self._market_id.get(
                        f"{pos_data.symbol}_{inst_type_suffix}"
                    )

                    if not symbol:
                        self._log.warning(
                            f"Symbol {pos_data.symbol} not found in market mapping"
                        )
                        continue

                    # Convert position side string to BitgetPositionSide enum
                    hold_side = BitgetPositionSide(pos_data.posSide)

                    # Convert to signed amount based on position side
                    signed_amount = Decimal(pos_data.total)
                    if hold_side == BitgetPositionSide.SHORT:
                        signed_amount = -signed_amount

                    # Parse position side
                    position_side = hold_side.parse_to_position_side()

                    # Create Position object
                    position = Position(
                        symbol=symbol,
                        exchange=self._exchange_id,
                        signed_amount=signed_amount,
                        entry_price=float(pos_data.avgPrice),
                        side=position_side,
                        unrealized_pnl=float(pos_data.unrealisedPnl),
                        realized_pnl=float(pos_data.curRealisedPnl),
                    )

                    # Apply position to cache
                    self._cache._apply_position(position)
                    self._log.debug(f"Initialized UTA position: {str(position)}")

        else:
            product_types = ["USDT-FUTURES", "USDC-FUTURES", "COIN-FUTURES"]

            for product_type in product_types:
                # Get all positions for this product type
                response = self._api_client.get_api_v2_mix_position_all_position(
                    productType=product_type
                )

                for pos_data in response.data:
                    # Check position mode - only support one-way mode
                    if pos_data.posMode != "one_way_mode":
                        raise PositionModeError(
                            f"Only one-way position mode is supported. Current mode: {pos_data.posMode}"
                        )

                    # Skip positions with zero amount
                    if float(pos_data.total) == 0:
                        continue

                    # Map symbol from Bitget format to our internal format
                    inst_type = BitgetInstType(product_type)
                    inst_type_suffix = self._inst_type_suffix(inst_type)
                    symbol = self._market_id.get(
                        f"{pos_data.symbol}_{inst_type_suffix}"
                    )

                    if not symbol:
                        self._log.warning(
                            f"Symbol {pos_data.symbol} not found in market mapping"
                        )
                        continue

                    # Convert holdSide string to BitgetPositionSide enum
                    hold_side = BitgetPositionSide(pos_data.holdSide)

                    # Convert to signed amount based on position side
                    signed_amount = Decimal(pos_data.total)
                    if hold_side == BitgetPositionSide.SHORT:
                        signed_amount = -signed_amount

                    # Parse position side
                    position_side = hold_side.parse_to_position_side()

                    # Create Position object
                    position = Position(
                        symbol=symbol,
                        exchange=self._exchange_id,
                        signed_amount=signed_amount,
                        entry_price=float(pos_data.openPriceAvg),
                        side=position_side,
                        unrealized_pnl=float(pos_data.unrealizedPL),
                        realized_pnl=float(pos_data.achievedProfits),
                    )

                    # Apply position to cache
                    self._cache._apply_position(position)
                    self._log.debug(f"Initialized position: {str(position)}")

    def _position_mode_check(self):
        """Check the position mode"""
        # Position mode check is performed in _init_position for each position
        # Only one-way mode is supported
        pass

    async def create_tp_sl_order(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce | None = TimeInForce.GTC,
        tp_order_type: OrderType | None = None,
        tp_trigger_price: Decimal | None = None,
        tp_price: Decimal | None = None,
        tp_trigger_type: TriggerType | None = TriggerType.LAST_PRICE,
        sl_order_type: OrderType | None = None,
        sl_trigger_price: Decimal | None = None,
        sl_price: Decimal | None = None,
        sl_trigger_type: TriggerType | None = TriggerType.LAST_PRICE,
        **kwargs,
    ) -> Order:
        """Create a take profit and stop loss order"""
        raise NotImplementedError

    def _price_to_precision(
        self,
        symbol: str,
        price: float,
        mode: Literal["round", "ceil", "floor"] = "round",
    ) -> Decimal:
        market = self._market[symbol]
        if market.spot:
            return super()._price_to_precision(symbol, price, mode)
        else:
            price: Decimal = Decimal(str(price))
            price_multiplier = Decimal(market.info.priceEndStep)
            multiplier_count = price / price_multiplier

            if mode == "round":
                price = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_HALF_UP)
                ) * price_multiplier
            elif mode == "ceil":
                price = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_CEILING)
                ) * price_multiplier
            elif mode == "floor":
                price = (
                    multiplier_count.quantize(Decimal("1"), rounding=ROUND_FLOOR)
                ) * price_multiplier
            return price

    async def create_order_ws(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal,
        time_in_force: TimeInForce = TimeInForce.GTC,
        reduce_only: bool = False,
        **kwargs,
    ):
        """Create an order"""
        self._registry.register_tmp_order(
            order=Order(
                oid=oid,
                exchange=self._exchange_id,
                symbol=symbol,
                status=OrderStatus.INITIALIZED,
                amount=amount,
                type=type,
                side=side,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                timestamp=self._clock.timestamp_ms(),
                reduce_only=reduce_only,
            )
        )
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} not found in market data.")

        if self._account_type.is_uta:
            category: str = kwargs.pop("category", self._get_inst_type(market))
            params = {
                "category": category.lower(),
                "symbol": market.id,
                "side": BitgetEnumParser.to_bitget_order_side(side).value,
                "qty": str(amount),
                "clientOid": oid,
            }
            if type.is_limit:
                params["price"] = str(price)
                params["timeInForce"] = BitgetEnumParser.to_bitget_time_in_force(
                    time_in_force
                ).value
                params["orderType"] = "limit"
            elif type.is_post_only:
                params["price"] = str(price)
                params["timeInForce"] = "post_only"
                params["orderType"] = "limit"
            elif type.is_market:
                bookl1 = self._cache.bookl1(symbol)
                if not bookl1:
                    raise ValueError(
                        "Please Subscribe to bookl1 first, since market order requires bookl1 data"
                    )
                if side.is_buy:
                    price = self._price_to_precision(
                        symbol, bookl1.ask * (1 + self._max_slippage), mode="ceil"
                    )
                else:
                    price = self._price_to_precision(
                        symbol, bookl1.bid * (1 - self._max_slippage), mode="floor"
                    )
                params["price"] = str(price)
                params["timeInForce"] = BitgetEnumParser.to_bitget_time_in_force(
                    TimeInForce.IOC
                ).value
                params["orderType"] = "limit"

            if reduce_only:
                params["reduceOnly"] = "yes"

            params.update(kwargs)

            await self._ws_api_client.uta_place_order(id=oid, **params)

        else:
            params = {
                "instId": market.id,
                "side": BitgetEnumParser.to_bitget_order_side(side).value,
                "size": str(amount),
                "clientOid": oid,
            }
            if type.is_limit:
                params["price"] = str(price)
                params["force"] = BitgetEnumParser.to_bitget_time_in_force(
                    time_in_force
                ).value
                params["orderType"] = "limit"
            elif type.is_post_only:
                params["price"] = str(price)
                params["force"] = "post_only"
                params["orderType"] = "limit"
            elif type.is_market:
                bookl1 = self._cache.bookl1(symbol)
                if not bookl1:
                    raise ValueError(
                        "Please Subscribe to bookl1 first, since market order requires bookl1 data"
                    )
                if side.is_buy:
                    price = self._price_to_precision(
                        symbol, bookl1.ask * (1 + self._max_slippage), mode="ceil"
                    )
                else:
                    price = self._price_to_precision(
                        symbol, bookl1.bid * (1 - self._max_slippage), mode="floor"
                    )
                params["price"] = str(price)
                params["force"] = BitgetEnumParser.to_bitget_time_in_force(
                    TimeInForce.IOC
                ).value
                params["orderType"] = "limit"

            params.update(kwargs)

            if market.swap:
                params["marginCoin"] = market.quote
                params["marginMode"] = kwargs.get("marginMode", "crossed")
                params["instType"] = self._get_inst_type(market)
                if reduce_only:
                    params["reduceOnly"] = "YES"
                await self._ws_api_client.future_place_order(id=oid, **params)
            else:
                await self._ws_api_client.spot_place_order(id=oid, **params)

    async def cancel_order_ws(self, oid: str, symbol: str, **kwargs):
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        instId = market.id
        if self._account_type.is_uta:
            await self._ws_api_client.uta_cancel_order(id=oid, clientOid=oid)
        else:
            if market.swap:
                await self._ws_api_client.future_cancel_order(
                    id=oid,
                    instId=instId,
                    clientOid=oid,
                    instType=self._get_inst_type(market),
                )
            else:
                await self._ws_api_client.spot_cancel_order(
                    id=oid, instId=instId, clientOid=oid
                )

    async def create_order(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal,
        time_in_force: TimeInForce = TimeInForce.GTC,
        reduce_only: bool = False,
        **kwargs,
    ) -> Order:
        """Create an order"""
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} not found in market data.")

        if self._account_type.is_uta:
            category = kwargs.pop("category", self._get_inst_type(market))
            params = {
                "category": category,
                "symbol": market.id,
                "qty": str(amount),
                "clientOid": oid,
            }
            if type.is_limit:
                params["price"] = str(price)
                params["timeInForce"] = BitgetEnumParser.to_bitget_time_in_force(
                    time_in_force
                ).value
                params["orderType"] = "limit"
            elif type.is_post_only:
                params["price"] = str(price)
                params["timeInForce"] = "post_only"
                params["orderType"] = "limit"
            elif type.is_market:
                bookl1 = self._cache.bookl1(symbol)
                if not bookl1:
                    raise ValueError(
                        "Please Subscribe to bookl1 first, since market order requires bookl1 data"
                    )
                if side.is_buy:
                    price = self._price_to_precision(
                        symbol, bookl1.ask * (1 + self._max_slippage), mode="ceil"
                    )
                else:
                    price = self._price_to_precision(
                        symbol, bookl1.bid * (1 - self._max_slippage), mode="floor"
                    )
                params["price"] = str(price)
                params["timeInForce"] = BitgetEnumParser.to_bitget_time_in_force(
                    TimeInForce.IOC
                ).value
                params["orderType"] = "limit"

            if reduce_only:
                params["reduceOnly"] = "yes"

            params.update(kwargs)

            try:
                res = await self._api_client.post_api_v3_trade_place_order(**params)

                order = Order(
                    exchange=self._exchange_id,
                    symbol=symbol,
                    eid=str(res.data.orderId),
                    oid=oid,
                    status=OrderStatus.PENDING,
                    side=side,
                    type=type,
                    price=float(price),
                    amount=amount,
                    time_in_force=time_in_force,
                    filled=Decimal(0),
                    remaining=amount,
                    reduce_only=reduce_only,
                    timestamp=res.requestTime,
                )

            except Exception as e:
                error_msg = f"{e.__class__.__name__}: {str(e)}"
                self._log.error(
                    f"Error creating order: {error_msg} params: {str(params)}"
                )
                order = Order(
                    exchange=self._exchange_id,
                    symbol=symbol,
                    oid=oid,
                    status=OrderStatus.FAILED,
                    side=side,
                    type=type,
                    price=float(price),
                    amount=amount,
                    time_in_force=time_in_force,
                    filled=Decimal(0),
                    remaining=amount,
                    reduce_only=reduce_only,
                    timestamp=self._clock.timestamp_ms(),
                )
            self.order_status_update(order)
            return
        else:
            params = {
                "symbol": market.id,
                "side": BitgetEnumParser.to_bitget_order_side(side).value,
                "size": str(amount),
                "clientOid": oid,
            }
            if type.is_limit:
                params["price"] = str(price)
                params["force"] = BitgetEnumParser.to_bitget_time_in_force(
                    time_in_force
                ).value
                params["orderType"] = "limit"
            elif type.is_post_only:
                params["price"] = str(price)
                params["force"] = "post_only"
                params["orderType"] = "limit"
            elif type.is_market:
                bookl1 = self._cache.bookl1(symbol)
                if not bookl1:
                    raise ValueError(
                        "Please Subscribe to bookl1 first, since market order requires bookl1 data"
                    )
                if side.is_buy:
                    price = self._price_to_precision(
                        symbol, bookl1.ask * (1 + self._max_slippage), mode="ceil"
                    )
                else:
                    price = self._price_to_precision(
                        symbol, bookl1.bid * (1 - self._max_slippage), mode="floor"
                    )
                params["price"] = str(price)
                params["force"] = BitgetEnumParser.to_bitget_time_in_force(
                    TimeInForce.IOC
                ).value
                params["orderType"] = "limit"

            params.update(kwargs)

            try:
                if market.swap:
                    params["marginCoin"] = market.quote
                    params["marginMode"] = kwargs.get("marginMode", "crossed")
                    params["productType"] = self._get_inst_type(market)
                    if reduce_only:
                        params["reduceOnly"] = "YES"

                    res = await self._api_client.post_api_v2_mix_order_place_order(
                        **params
                    )
                else:
                    res = await self._api_client.post_api_v2_spot_trade_place_order(
                        **params
                    )

                order = Order(
                    exchange=self._exchange_id,
                    symbol=symbol,
                    oid=oid,
                    eid=str(res.data.orderId),
                    status=OrderStatus.PENDING,
                    side=side,
                    type=type,
                    price=float(price),
                    amount=amount,
                    time_in_force=time_in_force,
                    filled=Decimal(0),
                    remaining=amount,
                    reduce_only=reduce_only,
                    timestamp=res.requestTime,
                )

            except Exception as e:
                error_msg = f"{e.__class__.__name__}: {str(e)}"
                self._log.error(
                    f"Error creating order: {error_msg} params: {str(params)}"
                )
                order = Order(
                    exchange=self._exchange_id,
                    symbol=symbol,
                    oid=oid,
                    status=OrderStatus.FAILED,
                    side=side,
                    type=type,
                    price=float(price),
                    amount=amount,
                    time_in_force=time_in_force,
                    filled=Decimal(0),
                    remaining=amount,
                    reduce_only=reduce_only,
                    timestamp=self._clock.timestamp_ms(),
                )
            self.order_status_update(order)

    async def create_batch_orders(
        self,
        orders: List[BatchOrderSubmit],
    ) -> List[Order]:
        """Create multiple orders in a batch"""
        raise NotImplementedError

    async def cancel_order(self, oid: str, symbol: str, **kwargs) -> Order:
        """Cancel an order"""
        market = self._market.get(symbol)
        if not market:
            raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
        id = market.id
        params = {
            "symbol": id,
            "clientOid": oid,
        }
        params.update(kwargs)
        try:
            if self._account_type.is_uta:
                res = await self._api_client.post_api_v3_trade_cancel_order(
                    clientOid=oid,
                )

            else:
                if market.swap:
                    params["productType"] = self._get_inst_type(market)
                    res = await self._api_client.post_api_v2_mix_order_cancel_order(
                        **params
                    )
                else:
                    res = await self._api_client.post_api_v2_spot_trade_cancel_order(
                        **params
                    )
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                eid=res.data.orderId,
                timestamp=res.requestTime,
                symbol=symbol,
                status=OrderStatus.CANCELING,
            )
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error canceling order: {error_msg} params: {str(params)}")
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=symbol,
                status=OrderStatus.CANCEL_FAILED,
            )
        self.order_status_update(order)

    async def modify_order(
        self,
        oid: str,
        symbol: str,
        order_id: str,
        side: OrderSide | None = None,
        price: Decimal | None = None,
        amount: Decimal | None = None,
        **kwargs,
    ) -> Order:
        """Modify an order"""
        raise NotImplementedError

    async def cancel_all_orders(self, symbol: str) -> bool:
        """Cancel all orders"""
        if not self._account_type.is_uta:
            raise NotImplementedError("Only UTA account type is supported")

        try:
            market = self._market.get(symbol)
            if not market:
                raise ValueError(f"Symbol {symbol} formated wrongly, or not supported")
            symbol = market.id
            category = self._get_inst_type(market)

            await self._api_client.post_api_v3_trade_cancel_symbol_order(
                symbol=symbol, category=category
            )

        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)} params: symbol={symbol} category={category}"
            self._log.error(f"Error canceling all orders: {error_msg}")
            return False

    def _ws_uta_msg_handler(self, raw: bytes):
        if raw == b"pong":
            self._ws_client._transport.notify_user_specific_pong_received()
            self._log.debug(f"Pong received: `{raw.decode()}`")
            return
        try:
            ws_msg = self._ws_msg_uta_general_decoder.decode(raw)
            if ws_msg.is_event_data:
                self._handle_event_data(ws_msg)
            elif ws_msg.arg.topic == "order":
                self._handle_uta_order_event(raw)
            elif ws_msg.arg.topic == "position":
                self._handle_uta_position_event(raw)
            elif ws_msg.arg.topic == "account":
                self._handle_uta_account_event(raw)

        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding message: {str(raw)} {e}")

    def _handle_uta_account_event(self, raw: bytes):
        msg = self._ws_msg_uta_account_decoder.decode(raw)
        balances = msg.parse_to_balances()
        self._cache._apply_balance(
            account_type=self._account_type,
            balances=balances,
        )
        for balance in balances:
            self._log.debug(
                f"Balance update: {balance.asset} - {balance.free} free, {balance.locked} locked"
            )

    def _handle_uta_position_event(self, raw: bytes):
        msg = self._ws_msg_uta_positions_decoder.decode(raw)
        for data in msg.data:
            # Determine suffix based on marginCoin
            if data.marginCoin in ["USDT", "USDC"]:
                suffix = "linear"
            else:
                suffix = "inverse"

            sym_id = f"{data.symbol}_{suffix}"
            symbol = self._market_id.get(sym_id)

            if not symbol:
                self._log.warning(f"Symbol not found for UTA position: {sym_id}")
                continue

            # Parse position side
            position_side = data.posSide.parse_to_position_side()
            signed_amount = Decimal(data.size)
            if position_side.is_short:
                signed_amount = -signed_amount

            # Create Position object
            position = Position(
                symbol=symbol,
                exchange=self._exchange_id,
                signed_amount=signed_amount,
                entry_price=float(data.openPriceAvg or 0.0),
                side=position_side,
                unrealized_pnl=float(data.unrealisedPnl or 0.0),
                realized_pnl=float(data.curRealisedPnl or 0.0),
            )

            # Apply position to cache
            self._cache._apply_position(position)
            self._log.debug(f"Position update: {str(position)}")

    def _handle_uta_order_event(self, raw: bytes):
        msg = self._ws_msg_uta_orders_decoder.decode(raw)
        self._log.debug(f"Received UTA order event: {str(msg)}")
        for data in msg.data:
            status = BitgetEnumParser.parse_order_status(data.orderStatus)
            if not status:
                continue
            # Build symbol ID using the helper method
            inst_type_suffix = self._uta_inst_type_suffix(data.category)
            sym_id = f"{data.symbol}_{inst_type_suffix}"

            symbol = self._market_id.get(sym_id)
            if not symbol:
                self._log.warning(f"Symbol not found for {sym_id}")
                continue

            # Parse order data
            timestamp = int(data.updatedTime)

            # Parse order type
            if data.orderType.is_market:
                order_type = OrderType.MARKET
            elif data.timeInForce.is_post_only:
                order_type = OrderType.POST_ONLY
            else:
                order_type = OrderType.LIMIT  # Default fallback

            # Calculate remaining quantity
            remaining = Decimal(data.qty) - Decimal(data.cumExecQty)

            # Calculate average price
            average_price = data.avgPrice or 0
            price = data.price or 0

            # Calculate fee
            total_fee = Decimal(0)
            fee_currency = None
            if data.feeDetail:
                total_fee = sum(Decimal(fee.fee) for fee in data.feeDetail)
                if data.feeDetail:
                    fee_currency = data.feeDetail[0].feeCoin

            order = Order(
                exchange=self._exchange_id,
                eid=data.orderId,
                oid=data.clientOid,
                timestamp=timestamp,
                symbol=symbol,
                type=order_type,
                side=BitgetEnumParser.parse_order_side(data.side),
                price=price,
                average=average_price,
                amount=Decimal(data.qty),
                filled=Decimal(data.cumExecQty),
                remaining=remaining,
                status=status,
                fee=total_fee,
                fee_currency=fee_currency,
                cum_cost=Decimal(data.cumExecValue or 0),
                reduce_only=data.reduceOnly == "yes",
            )
            self._log.debug(f"Order update: {str(order)}")
            self.order_status_update(order)

    def _ws_msg_handler(self, raw: bytes):
        """Handle incoming WebSocket messages"""
        # Process the message based on its type
        if raw == b"pong":
            self._ws_client._transport.notify_user_specific_pong_received()
            self._log.debug(f"Pong received: `{raw.decode()}`")
            return

        try:
            ws_msg = self._ws_msg_general_decoder.decode(raw)
            if ws_msg.is_event_data:
                self._handle_event_data(ws_msg)
            elif ws_msg.arg.channel == "orders":
                self._handle_orders_event(raw, ws_msg.arg)
            elif ws_msg.arg.channel == "positions":
                self._handle_positions_event(raw, ws_msg.arg)
            elif ws_msg.arg.channel == "account":
                self._handle_account_event(raw, ws_msg.arg)

        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding message: {str(raw)} {e}")

    def _handle_account_event(self, raw: bytes, arg: BitgetWsArgMsg):
        if arg.instType.is_spot:
            msg = self._ws_msg_spot_account_decoder.decode(raw)
        else:
            msg = self._ws_msg_futures_account_decoder.decode(raw)
        self._cache._apply_balance(
            account_type=self._account_type, balances=msg.parse_to_balances()
        )

    def _handle_positions_event(self, raw: bytes, arg: BitgetWsArgMsg):
        msg = self._ws_msg_positions_decoder.decode(raw)

        # Get existing positions for this specific instrument type
        existing_positions = self._cache.get_all_positions(exchange=self._exchange_id)
        inst_type_suffix = self._inst_type_suffix(arg.instType)

        # Filter existing positions to only include those from this instrument type
        if arg.instType.is_usdt_swap:
            existing_positions_for_inst_type = {
                symbol: pos
                for symbol, pos in existing_positions.items()
                if self._market[symbol].quote == "USDT"
            }
        elif arg.instType.is_usdc_swap:
            existing_positions_for_inst_type = {
                symbol: pos
                for symbol, pos in existing_positions.items()
                if self._market[symbol].quote == "USDC"
            }
        elif arg.instType.is_inverse:
            existing_positions_for_inst_type = {
                symbol: pos
                for symbol, pos in existing_positions.items()
                if self._market[symbol].inverse
            }

        active_symbols = set()

        for data in msg.data:
            sym_id = data.instId
            symbol = self._market_id[f"{sym_id}_{inst_type_suffix}"]
            active_symbols.add(symbol)

            # Convert Bitget position data to Position
            signed_amount = Decimal(data.total)
            if data.holdSide.is_short:
                signed_amount = -signed_amount

            position_side = data.holdSide.parse_to_position_side()

            position = Position(
                symbol=symbol,
                exchange=self._exchange_id,
                signed_amount=signed_amount,
                entry_price=float(data.openPriceAvg),
                side=position_side,
                unrealized_pnl=float(data.unrealizedPL),
                realized_pnl=float(data.achievedProfits),
            )

            self._cache._apply_position(position)
            self._log.debug(f"Position update: {str(position)}")

        # Close positions that are not in the snapshot (position goes to 0)
        # Only check positions for the current instrument type
        for symbol, existing_position in existing_positions_for_inst_type.items():
            if symbol not in active_symbols:
                # Create a closed position with signed_amount = 0
                closed_position = Position(
                    symbol=symbol,
                    exchange=self._exchange_id,
                    signed_amount=Decimal("0"),
                    entry_price=existing_position.entry_price,
                    side=None,
                    unrealized_pnl=0,
                    realized_pnl=existing_position.realized_pnl,
                )
                self._cache._apply_position(closed_position)
                self._log.debug(f"Position closed: {str(closed_position)}")

    def _handle_event_data(self, msg: BitgetWsGeneralMsg | BitgetWsUtaGeneralMsg):
        if msg.event == "subscribe":
            arg = msg.arg
            self._log.debug(f"Subscribed to {arg.message}")
        elif msg.event == "error":
            code = msg.code
            error_msg = msg.msg
            self._log.error(f"Subscribed error code={code} {error_msg}")
        elif msg.event == "login":
            if msg.code == 0:
                self._log.debug("WebSocket login successful")
            else:
                self._log.error(f"WebSocket login failed: {msg.msg}")

    def _handle_orders_event(self, raw: bytes, arg: BitgetWsArgMsg):
        msg = self._ws_msg_orders_decoder.decode(raw)
        self._log.debug(f"Received order event: {str(msg)}")
        for data in msg.data:
            sym_id = data.instId
            timestamp = int(data.uTime)
            if data.orderType.is_market:
                typ = OrderType.MARKET
            elif data.force.is_post_only:
                typ = OrderType.POST_ONLY
            else:
                typ = OrderType.LIMIT

            status = BitgetEnumParser.parse_order_status(data.status)
            side = BitgetEnumParser.parse_order_side(data.side)

            if fee_details := data.feeDetail:
                fee = fee_details[0].fee
                fee_currency = fee_details[0].feeCoin
            else:
                fee = 0
                fee_currency = None
            filled = data.accBaseVolume
            last_filled = data.accBaseVolume or 0
            last_filled_price = data.fillPrice or 0
            price = data.price or 0
            average = data.priceAvg or 0

            if arg.instType.is_spot:
                symbol = self._market_id[f"{sym_id}_spot"]
                order = Order(
                    exchange=self._exchange_id,
                    symbol=symbol,
                    eid=str(data.orderId),
                    oid=str(data.clientOid),
                    amount=Decimal(data.newSize),
                    filled=Decimal(filled),
                    timestamp=timestamp,
                    status=status,
                    side=side,
                    type=typ,
                    last_filled=Decimal(last_filled),
                    last_filled_price=float(last_filled_price),
                    remaining=Decimal(data.newSize) - Decimal(data.accBaseVolume),
                    fee=Decimal(fee),
                    fee_currency=fee_currency,
                    price=float(price),
                    average=float(average),
                    cost=Decimal(last_filled) * Decimal(last_filled_price),
                    cum_cost=Decimal(filled) * Decimal(average),
                )
            else:
                symbol = self._market_id[
                    f"{sym_id}_{self._inst_type_suffix(arg.instType)}"
                ]
                order = Order(
                    exchange=self._exchange_id,
                    symbol=symbol,
                    eid=str(data.orderId),
                    oid=str(data.clientOid),
                    amount=Decimal(data.size),
                    filled=Decimal(filled),
                    timestamp=timestamp,
                    status=status,
                    side=side,
                    type=typ,
                    last_filled=Decimal(last_filled),
                    last_filled_price=float(last_filled_price),
                    remaining=Decimal(data.size) - Decimal(data.accBaseVolume),
                    fee=Decimal(fee),
                    fee_currency=fee_currency,
                    average=float(average),
                    price=float(price),
                    cost=Decimal(last_filled) * Decimal(last_filled_price),
                    cum_cost=Decimal(filled) * Decimal(average),
                    reduce_only=data.reduceOnly == "yes",
                )
            self.order_status_update(order)



import asyncio
from decimal import Decimal
from typing import Dict, List
from nexustrader.constants import AccountType, SubmitType
from nexustrader.schema import OrderSubmit, InstrumentId
from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import MessageBus, LiveClock
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.exchange.binance import BinanceAccountType
from nexustrader.exchange.binance.schema import BinanceMarket
from nexustrader.base import ExecutionManagementSystem


class BinanceExecutionManagementSystem(ExecutionManagementSystem):
    _market: Dict[str, BinanceMarket]

    BINANCE_SPOT_PRIORITY = [
        BinanceAccountType.ISOLATED_MARGIN,
        BinanceAccountType.MARGIN,
        BinanceAccountType.SPOT_TESTNET,
        BinanceAccountType.SPOT,
    ]

    def __init__(
        self,
        market: Dict[str, BinanceMarket],
        cache: AsyncCache,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        registry: OrderRegistry,
        is_mock: bool = False,
    ):
        super().__init__(
            market=market,
            cache=cache,
            msgbus=msgbus,
            clock=clock,
            task_manager=task_manager,
            registry=registry,
            is_mock=is_mock,
        )
        self._binance_spot_account_type: BinanceAccountType = None
        self._binance_linear_account_type: BinanceAccountType = None
        self._binance_inverse_account_type: BinanceAccountType = None
        self._binance_pm_account_type: BinanceAccountType = None

    def _set_account_type(self):
        account_types = self._private_connectors.keys()

        if self._is_mock:
            self._binance_spot_account_type = BinanceAccountType.SPOT_MOCK
            self._binance_linear_account_type = BinanceAccountType.LINEAR_MOCK
            self._binance_inverse_account_type = BinanceAccountType.INVERSE_MOCK
        else:
            if BinanceAccountType.PORTFOLIO_MARGIN in account_types:
                self._binance_pm_account_type = BinanceAccountType.PORTFOLIO_MARGIN
                return

            for account_type in self.BINANCE_SPOT_PRIORITY:
                if account_type in account_types:
                    self._binance_spot_account_type = account_type
                    break

            self._binance_linear_account_type = (
                BinanceAccountType.USD_M_FUTURE_TESTNET
                if BinanceAccountType.USD_M_FUTURE_TESTNET in account_types
                else BinanceAccountType.USD_M_FUTURE
            )

            self._binance_inverse_account_type = (
                BinanceAccountType.COIN_M_FUTURE_TESTNET
                if BinanceAccountType.COIN_M_FUTURE_TESTNET in account_types
                else BinanceAccountType.COIN_M_FUTURE
            )

    def _instrument_id_to_account_type(
        self, instrument_id: InstrumentId
    ) -> AccountType:
        if self._binance_pm_account_type:
            return self._binance_pm_account_type
        if instrument_id.is_spot:
            return self._binance_spot_account_type
        elif instrument_id.is_linear:
            return self._binance_linear_account_type
        elif instrument_id.is_inverse:
            return self._binance_inverse_account_type

    def _build_order_submit_queues(self):
        for account_type in self._private_connectors.keys():
            if isinstance(account_type, BinanceAccountType):
                self._order_submit_queues[account_type] = asyncio.Queue()

    def _submit_order(
        self,
        order: OrderSubmit | List[OrderSubmit],
        submit_type: SubmitType,
        account_type: AccountType | None = None,
    ):
        if isinstance(order, list):
            if not account_type:
                account_type = self._instrument_id_to_account_type(
                    order[0].instrument_id
                )

            # Split batch orders into chunks of 5
            for i in range(0, len(order), 5):
                batch = order[i : i + 5]
                self._order_submit_queues[account_type].put_nowait((batch, submit_type))
        else:
            if not account_type:
                account_type = self._instrument_id_to_account_type(order.instrument_id)
            self._order_submit_queues[account_type].put_nowait((order, submit_type))

    def _get_min_order_amount(self, symbol: str, market: BinanceMarket) -> Decimal:
        book = self._cache.bookl1(symbol)
        cost_min = market.limits.cost.min
        amount_min = market.limits.amount.min

        min_order_amount = max(cost_min * 1.01 / book.mid, amount_min)
        min_order_amount = self._amount_to_precision(
            symbol, min_order_amount, mode="ceil"
        )
        return min_order_amount
